"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2014 The Chromium Authors. All rights reserved.
Use of this source code is governed by a BSD-style license that can be
found in the LICENSE file.

Request and reply to the "checkin server" devices poll every few hours.
"""

from . import android_checkin_pb2
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class GservicesSetting(google.protobuf.message.Message):
    """A concrete name/value pair sent to the device's Gservices database."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    VALUE_FIELD_NUMBER: builtins.int
    name: builtins.bytes
    value: builtins.bytes
    def __init__(
        self,
        *,
        name: builtins.bytes | None = ...,
        value: builtins.bytes | None = ...,
    ) -> None: ...
    def HasField(
        self, field_name: typing_extensions.Literal["name", b"name", "value", b"value"]
    ) -> builtins.bool: ...
    def ClearField(
        self, field_name: typing_extensions.Literal["name", b"name", "value", b"value"]
    ) -> None: ...

global___GservicesSetting = GservicesSetting

@typing_extensions.final
class AndroidCheckinRequest(google.protobuf.message.Message):
    """Devices send this every few hours to tell us how they're doing."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    IMEI_FIELD_NUMBER: builtins.int
    MEID_FIELD_NUMBER: builtins.int
    MAC_ADDR_FIELD_NUMBER: builtins.int
    MAC_ADDR_TYPE_FIELD_NUMBER: builtins.int
    SERIAL_NUMBER_FIELD_NUMBER: builtins.int
    ESN_FIELD_NUMBER: builtins.int
    ID_FIELD_NUMBER: builtins.int
    LOGGING_ID_FIELD_NUMBER: builtins.int
    DIGEST_FIELD_NUMBER: builtins.int
    LOCALE_FIELD_NUMBER: builtins.int
    CHECKIN_FIELD_NUMBER: builtins.int
    DESIRED_BUILD_FIELD_NUMBER: builtins.int
    MARKET_CHECKIN_FIELD_NUMBER: builtins.int
    ACCOUNT_COOKIE_FIELD_NUMBER: builtins.int
    TIME_ZONE_FIELD_NUMBER: builtins.int
    SECURITY_TOKEN_FIELD_NUMBER: builtins.int
    VERSION_FIELD_NUMBER: builtins.int
    OTA_CERT_FIELD_NUMBER: builtins.int
    FRAGMENT_FIELD_NUMBER: builtins.int
    USER_NAME_FIELD_NUMBER: builtins.int
    USER_SERIAL_NUMBER_FIELD_NUMBER: builtins.int
    imei: builtins.str
    """IMEI (used by GSM phones) is sent and stored as 15 decimal
    digits; the 15th is a check digit.
    IMEI, reported but not logged.
    """
    meid: builtins.str
    """MEID (used by CDMA phones) is sent and stored as 14 hexadecimal
    digits (no check digit).
    MEID, reported but not logged.
    """
    @property
    def mac_addr(
        self,
    ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """MAC address (used by non-phone devices).  12 hexadecimal digits;
        no separators (eg "0016E6513AC2", not "00:16:E6:51:3A:C2").
        MAC address, reported but not logged.
        """
    @property
    def mac_addr_type(
        self,
    ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """An array parallel to mac_addr, describing the type of interface.
        Currently accepted values: "wifi", "ethernet", "bluetooth".  If
        not present, "wifi" is assumed.
        """
    serial_number: builtins.str
    """Serial number (a manufacturer-defined unique hardware
    identifier).  Alphanumeric, case-insensitive.
    """
    esn: builtins.str
    """Older CDMA networks use an ESN (8 hex digits) instead of an MEID.
    ESN, reported but not logged
    """
    id: builtins.int
    """Android device ID, not logged"""
    logging_id: builtins.int
    """Pseudonymous logging ID for Sawmill"""
    digest: builtins.str
    """Digest of device provisioning, not logged."""
    locale: builtins.str
    """Current locale in standard (xx_XX) format"""
    @property
    def checkin(self) -> android_checkin_pb2.AndroidCheckinProto: ...
    desired_build: builtins.str
    """DEPRECATED, see AndroidCheckinProto.requested_group"""
    market_checkin: builtins.str
    """Blob of data from the Market app to be passed to Market API server"""
    @property
    def account_cookie(
        self,
    ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """SID cookies of any google accounts stored on the phone.  Not logged."""
    time_zone: builtins.str
    """Time zone.  Not currently logged."""
    security_token: builtins.int
    """Security token used to validate the checkin request.
    Required for android IDs issued to Froyo+ devices, not for legacy IDs.
    """
    version: builtins.int
    """Version of checkin protocol.

    There are currently two versions:

    - version field missing: android IDs are assigned based on
      hardware identifiers.  unsecured in the sense that you can
      "unregister" someone's phone by sending a registration request
      with their IMEI/MEID/MAC.

    - version=2: android IDs are assigned randomly.  The device is
      sent a security token that must be included in all future
      checkins for that android id.

    - version=3: same as version 2, but the 'fragment' field is
      provided, and the device understands incremental updates to the
      gservices table (ie, only returning the keys whose values have
      changed.)

    (version=1 was skipped to avoid confusion with the "missing"
    version field that is effectively version 1.)
    """
    @property
    def ota_cert(
        self,
    ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """OTA certs accepted by device (base-64 SHA-1 of cert files).  Not
        logged.
        """
    fragment: builtins.int
    """Honeycomb and newer devices send configuration data with their checkin.
    optional DeviceConfigurationProto device_configuration = 18;

    A single CheckinTask on the device may lead to multiple checkin
    requests if there is too much log data to upload in a single
    request.  For version 3 and up, this field will be filled in with
    the number of the request, starting with 0.
    """
    user_name: builtins.str
    """For devices supporting multiple users, the name of the current
    profile (they all check in independently, just as if they were
    multiple physical devices).  This may not be set, even if the
    device is using multiuser.  (checkin.user_number should be set to
    the ordinal of the user.)
    """
    user_serial_number: builtins.int
    """For devices supporting multiple user profiles, the serial number
    for the user checking in.  Not logged.  May not be set, even if
    the device supportes multiuser.  checkin.user_number is the
    ordinal of the user (0, 1, 2, ...), which may be reused if users
    are deleted and re-created.  user_serial_number is never reused
    (unless the device is wiped).
    """
    def __init__(
        self,
        *,
        imei: builtins.str | None = ...,
        meid: builtins.str | None = ...,
        mac_addr: collections.abc.Iterable[builtins.str] | None = ...,
        mac_addr_type: collections.abc.Iterable[builtins.str] | None = ...,
        serial_number: builtins.str | None = ...,
        esn: builtins.str | None = ...,
        id: builtins.int | None = ...,
        logging_id: builtins.int | None = ...,
        digest: builtins.str | None = ...,
        locale: builtins.str | None = ...,
        checkin: android_checkin_pb2.AndroidCheckinProto | None = ...,
        desired_build: builtins.str | None = ...,
        market_checkin: builtins.str | None = ...,
        account_cookie: collections.abc.Iterable[builtins.str] | None = ...,
        time_zone: builtins.str | None = ...,
        security_token: builtins.int | None = ...,
        version: builtins.int | None = ...,
        ota_cert: collections.abc.Iterable[builtins.str] | None = ...,
        fragment: builtins.int | None = ...,
        user_name: builtins.str | None = ...,
        user_serial_number: builtins.int | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "checkin",
            b"checkin",
            "desired_build",
            b"desired_build",
            "digest",
            b"digest",
            "esn",
            b"esn",
            "fragment",
            b"fragment",
            "id",
            b"id",
            "imei",
            b"imei",
            "locale",
            b"locale",
            "logging_id",
            b"logging_id",
            "market_checkin",
            b"market_checkin",
            "meid",
            b"meid",
            "security_token",
            b"security_token",
            "serial_number",
            b"serial_number",
            "time_zone",
            b"time_zone",
            "user_name",
            b"user_name",
            "user_serial_number",
            b"user_serial_number",
            "version",
            b"version",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "account_cookie",
            b"account_cookie",
            "checkin",
            b"checkin",
            "desired_build",
            b"desired_build",
            "digest",
            b"digest",
            "esn",
            b"esn",
            "fragment",
            b"fragment",
            "id",
            b"id",
            "imei",
            b"imei",
            "locale",
            b"locale",
            "logging_id",
            b"logging_id",
            "mac_addr",
            b"mac_addr",
            "mac_addr_type",
            b"mac_addr_type",
            "market_checkin",
            b"market_checkin",
            "meid",
            b"meid",
            "ota_cert",
            b"ota_cert",
            "security_token",
            b"security_token",
            "serial_number",
            b"serial_number",
            "time_zone",
            b"time_zone",
            "user_name",
            b"user_name",
            "user_serial_number",
            b"user_serial_number",
            "version",
            b"version",
        ],
    ) -> None: ...

global___AndroidCheckinRequest = AndroidCheckinRequest

@typing_extensions.final
class AndroidCheckinResponse(google.protobuf.message.Message):
    """The response to the device."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    STATS_OK_FIELD_NUMBER: builtins.int
    TIME_MSEC_FIELD_NUMBER: builtins.int
    DIGEST_FIELD_NUMBER: builtins.int
    SETTINGS_DIFF_FIELD_NUMBER: builtins.int
    DELETE_SETTING_FIELD_NUMBER: builtins.int
    SETTING_FIELD_NUMBER: builtins.int
    MARKET_OK_FIELD_NUMBER: builtins.int
    ANDROID_ID_FIELD_NUMBER: builtins.int
    SECURITY_TOKEN_FIELD_NUMBER: builtins.int
    VERSION_INFO_FIELD_NUMBER: builtins.int
    stats_ok: builtins.bool
    """Whether statistics were recorded properly."""
    time_msec: builtins.int
    """Time of day from server (Java epoch)."""
    digest: builtins.str
    """repeated AndroidIntentProto intent = 2;

    Provisioning is sent if the request included an obsolete digest.

    For version <= 2, 'digest' contains the digest that should be
    sent back to the server on the next checkin, and 'setting'
    contains the entire gservices table (which replaces the entire
    current table on the device).

    for version >= 3, 'digest' will be absent.  If 'settings_diff'
    is false, then 'setting' contains the entire table, as in version
    2.  If 'settings_diff' is true, then 'delete_setting' contains
    the keys to delete, and 'setting' contains only keys to be added
    or for which the value has changed.  All other keys in the
    current table should be left untouched.  If 'settings_diff' is
    absent, don't touch the existing gservices table.
    """
    settings_diff: builtins.bool
    @property
    def delete_setting(
        self,
    ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[
        builtins.str
    ]: ...
    @property
    def setting(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
        global___GservicesSetting
    ]: ...
    market_ok: builtins.bool
    """If Market got the market_checkin data OK."""
    android_id: builtins.int
    """From the request, or newly assigned"""
    security_token: builtins.int
    """The associated security token"""
    version_info: builtins.str
    """NEXT TAG: 12"""
    def __init__(
        self,
        *,
        stats_ok: builtins.bool | None = ...,
        time_msec: builtins.int | None = ...,
        digest: builtins.str | None = ...,
        settings_diff: builtins.bool | None = ...,
        delete_setting: collections.abc.Iterable[builtins.str] | None = ...,
        setting: collections.abc.Iterable[global___GservicesSetting] | None = ...,
        market_ok: builtins.bool | None = ...,
        android_id: builtins.int | None = ...,
        security_token: builtins.int | None = ...,
        version_info: builtins.str | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "android_id",
            b"android_id",
            "digest",
            b"digest",
            "market_ok",
            b"market_ok",
            "security_token",
            b"security_token",
            "settings_diff",
            b"settings_diff",
            "stats_ok",
            b"stats_ok",
            "time_msec",
            b"time_msec",
            "version_info",
            b"version_info",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "android_id",
            b"android_id",
            "delete_setting",
            b"delete_setting",
            "digest",
            b"digest",
            "market_ok",
            b"market_ok",
            "security_token",
            b"security_token",
            "setting",
            b"setting",
            "settings_diff",
            b"settings_diff",
            "stats_ok",
            b"stats_ok",
            "time_msec",
            b"time_msec",
            "version_info",
            b"version_info",
        ],
    ) -> None: ...

global___AndroidCheckinResponse = AndroidCheckinResponse



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mcs.proto
"""Generated protocol buffer code."""

from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\tmcs.proto\x12\tmcs_proto"S\n\rHeartbeatPing\x12\x11\n\tstream_id\x18\x01 \x01(\x05\x12\x1f\n\x17last_stream_id_received\x18\x02 \x01(\x05\x12\x0e\n\x06status\x18\x03 \x01(\x03"R\n\x0cHeartbeatAck\x12\x11\n\tstream_id\x18\x01 \x01(\x05\x12\x1f\n\x17last_stream_id_received\x18\x02 \x01(\x05\x12\x0e\n\x06status\x18\x03 \x01(\x03"a\n\tErrorInfo\x12\x0c\n\x04\x63ode\x18\x01 \x02(\x05\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x0c\n\x04type\x18\x03 \x01(\t\x12\'\n\textension\x18\x04 \x01(\x0b\x32\x14.mcs_proto.Extension"&\n\x07Setting\x12\x0c\n\x04name\x18\x01 \x02(\t\x12\r\n\x05value\x18\x02 \x02(\t"A\n\rHeartbeatStat\x12\n\n\x02ip\x18\x01 \x02(\t\x12\x0f\n\x07timeout\x18\x02 \x02(\x08\x12\x13\n\x0binterval_ms\x18\x03 \x02(\x05"G\n\x0fHeartbeatConfig\x12\x13\n\x0bupload_stat\x18\x01 \x01(\x08\x12\n\n\x02ip\x18\x02 \x01(\t\x12\x13\n\x0binterval_ms\x18\x03 \x01(\x05"\xdb\x02\n\x0b\x43lientEvent\x12)\n\x04type\x18\x01 \x01(\x0e\x32\x1b.mcs_proto.ClientEvent.Type\x12\x1f\n\x17number_discarded_events\x18\x64 \x01(\r\x12\x15\n\x0cnetwork_type\x18\xc8\x01 \x01(\x05\x12#\n\x1atime_connection_started_ms\x18\xca\x01 \x01(\x04\x12!\n\x18time_connection_ended_ms\x18\xcb\x01 \x01(\x04\x12\x13\n\nerror_code\x18\xcc\x01 \x01(\x05\x12\'\n\x1etime_connection_established_ms\x18\xac\x02 \x01(\x04"[\n\x04Type\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x14\n\x10\x44ISCARDED_EVENTS\x10\x01\x12\x15\n\x11\x46\x41ILED_CONNECTION\x10\x02\x12\x19\n\x15SUCCESSFUL_CONNECTION\x10\x03J\x06\x08\xc9\x01\x10\xca\x01"\xff\x03\n\x0cLoginRequest\x12\n\n\x02id\x18\x01 \x02(\t\x12\x0e\n\x06\x64omain\x18\x02 \x02(\t\x12\x0c\n\x04user\x18\x03 \x02(\t\x12\x10\n\x08resource\x18\x04 \x02(\t\x12\x12\n\nauth_token\x18\x05 \x02(\t\x12\x11\n\tdevice_id\x18\x06 \x01(\t\x12\x13\n\x0blast_rmq_id\x18\x07 \x01(\x03\x12#\n\x07setting\x18\x08 \x03(\x0b\x32\x12.mcs_proto.Setting\x12\x1e\n\x16received_persistent_id\x18\n \x03(\t\x12\x1a\n\x12\x61\x64\x61ptive_heartbeat\x18\x0c \x01(\x08\x12\x30\n\x0eheartbeat_stat\x18\r \x01(\x0b\x32\x18.mcs_proto.HeartbeatStat\x12\x10\n\x08use_rmq2\x18\x0e \x01(\x08\x12\x12\n\naccount_id\x18\x0f \x01(\x03\x12\x39\n\x0c\x61uth_service\x18\x10 \x01(\x0e\x32#.mcs_proto.LoginRequest.AuthService\x12\x14\n\x0cnetwork_type\x18\x11 \x01(\x05\x12\x0e\n\x06status\x18\x12 \x01(\x03\x12,\n\x0c\x63lient_event\x18\x16 \x03(\x0b\x32\x16.mcs_proto.ClientEvent"\x1d\n\x0b\x41uthService\x12\x0e\n\nANDROID_ID\x10\x02J\x04\x08\x13\x10\x14J\x04\x08\x14\x10\x15J\x04\x08\x15\x10\x16"\xf6\x01\n\rLoginResponse\x12\n\n\x02id\x18\x01 \x02(\t\x12\x0b\n\x03jid\x18\x02 \x01(\t\x12#\n\x05\x65rror\x18\x03 \x01(\x0b\x32\x14.mcs_proto.ErrorInfo\x12#\n\x07setting\x18\x04 \x03(\x0b\x32\x12.mcs_proto.Setting\x12\x11\n\tstream_id\x18\x05 \x01(\x05\x12\x1f\n\x17last_stream_id_received\x18\x06 \x01(\x05\x12\x34\n\x10heartbeat_config\x18\x07 \x01(\x0b\x32\x1a.mcs_proto.HeartbeatConfig\x12\x18\n\x10server_timestamp\x18\x08 \x01(\x03"/\n\x11StreamErrorStanza\x12\x0c\n\x04type\x18\x01 \x02(\t\x12\x0c\n\x04text\x18\x02 \x01(\t"\x07\n\x05\x43lose"%\n\tExtension\x12\n\n\x02id\x18\x01 \x02(\x05\x12\x0c\n\x04\x64\x61ta\x18\x02 \x02(\x0c"\xdd\x02\n\x08IqStanza\x12\x0e\n\x06rmq_id\x18\x01 \x01(\x03\x12(\n\x04type\x18\x02 \x02(\x0e\x32\x1a.mcs_proto.IqStanza.IqType\x12\n\n\x02id\x18\x03 \x02(\t\x12\x0c\n\x04\x66rom\x18\x04 \x01(\t\x12\n\n\x02to\x18\x05 \x01(\t\x12#\n\x05\x65rror\x18\x06 \x01(\x0b\x32\x14.mcs_proto.ErrorInfo\x12\'\n\textension\x18\x07 \x01(\x0b\x32\x14.mcs_proto.Extension\x12\x15\n\rpersistent_id\x18\x08 \x01(\t\x12\x11\n\tstream_id\x18\t \x01(\x05\x12\x1f\n\x17last_stream_id_received\x18\n \x01(\x05\x12\x12\n\naccount_id\x18\x0b \x01(\x03\x12\x0e\n\x06status\x18\x0c \x01(\x03"4\n\x06IqType\x12\x07\n\x03GET\x10\x00\x12\x07\n\x03SET\x10\x01\x12\n\n\x06RESULT\x10\x02\x12\x0c\n\x08IQ_ERROR\x10\x03"%\n\x07\x41ppData\x12\x0b\n\x03key\x18\x01 \x02(\t\x12\r\n\x05value\x18\x02 \x02(\t"\xf4\x02\n\x11\x44\x61taMessageStanza\x12\n\n\x02id\x18\x02 \x01(\t\x12\x0c\n\x04\x66rom\x18\x03 \x02(\t\x12\n\n\x02to\x18\x04 \x01(\t\x12\x10\n\x08\x63\x61tegory\x18\x05 \x02(\t\x12\r\n\x05token\x18\x06 \x01(\t\x12$\n\x08\x61pp_data\x18\x07 \x03(\x0b\x32\x12.mcs_proto.AppData\x12\x1b\n\x13\x66rom_trusted_server\x18\x08 \x01(\x08\x12\x15\n\rpersistent_id\x18\t \x01(\t\x12\x11\n\tstream_id\x18\n \x01(\x05\x12\x1f\n\x17last_stream_id_received\x18\x0b \x01(\x05\x12\x0e\n\x06reg_id\x18\r \x01(\t\x12\x16\n\x0e\x64\x65vice_user_id\x18\x10 \x01(\x03\x12\x0b\n\x03ttl\x18\x11 \x01(\x05\x12\x0c\n\x04sent\x18\x12 \x01(\x03\x12\x0e\n\x06queued\x18\x13 \x01(\x05\x12\x0e\n\x06status\x18\x14 \x01(\x03\x12\x10\n\x08raw_data\x18\x15 \x01(\x0c\x12\x15\n\rimmediate_ack\x18\x18 \x01(\x08"\x0b\n\tStreamAck"\x1a\n\x0cSelectiveAck\x12\n\n\x02id\x18\x01 \x03(\tB\x02H\x03'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "mcs_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    _globals["DESCRIPTOR"]._options = None
    _globals["DESCRIPTOR"]._serialized_options = b"H\003"
    _globals["_HEARTBEATPING"]._serialized_start = 24
    _globals["_HEARTBEATPING"]._serialized_end = 107
    _globals["_HEARTBEATACK"]._serialized_start = 109
    _globals["_HEARTBEATACK"]._serialized_end = 191
    _globals["_ERRORINFO"]._serialized_start = 193
    _globals["_ERRORINFO"]._serialized_end = 290
    _globals["_SETTING"]._serialized_start = 292
    _globals["_SETTING"]._serialized_end = 330
    _globals["_HEARTBEATSTAT"]._serialized_start = 332
    _globals["_HEARTBEATSTAT"]._serialized_end = 397
    _globals["_HEARTBEATCONFIG"]._serialized_start = 399
    _globals["_HEARTBEATCONFIG"]._serialized_end = 470
    _globals["_CLIENTEVENT"]._serialized_start = 473
    _globals["_CLIENTEVENT"]._serialized_end = 820
    _globals["_CLIENTEVENT_TYPE"]._serialized_start = 721
    _globals["_CLIENTEVENT_TYPE"]._serialized_end = 812
    _globals["_LOGINREQUEST"]._serialized_start = 823
    _globals["_LOGINREQUEST"]._serialized_end = 1334
    _globals["_LOGINREQUEST_AUTHSERVICE"]._serialized_start = 1287
    _globals["_LOGINREQUEST_AUTHSERVICE"]._serialized_end = 1316
    _globals["_LOGINRESPONSE"]._serialized_start = 1337
    _globals["_LOGINRESPONSE"]._serialized_end = 1583
    _globals["_STREAMERRORSTANZA"]._serialized_start = 1585
    _globals["_STREAMERRORSTANZA"]._serialized_end = 1632
    _globals["_CLOSE"]._serialized_start = 1634
    _globals["_CLOSE"]._serialized_end = 1641
    _globals["_EXTENSION"]._serialized_start = 1643
    _globals["_EXTENSION"]._serialized_end = 1680
    _globals["_IQSTANZA"]._serialized_start = 1683
    _globals["_IQSTANZA"]._serialized_end = 2032
    _globals["_IQSTANZA_IQTYPE"]._serialized_start = 1980
    _globals["_IQSTANZA_IQTYPE"]._serialized_end = 2032
    _globals["_APPDATA"]._serialized_start = 2034
    _globals["_APPDATA"]._serialized_end = 2071
    _globals["_DATAMESSAGESTANZA"]._serialized_start = 2074
    _globals["_DATAMESSAGESTANZA"]._serialized_end = 2446
    _globals["_STREAMACK"]._serialized_start = 2448
    _globals["_STREAMACK"]._serialized_end = 2459
    _globals["_SELECTIVEACK"]._serialized_start = 2461
    _globals["_SELECTIVEACK"]._serialized_end = 2487
# @@protoc_insertion_point(module_scope)



import logging
import asyncio
import os
import json
import string
import secrets

from .firebase_messaging import FcmPushClient, FcmRegisterConfig
from typing import Any


from .const import (
    FCM_PROJECT_ID,
    FCM_API_KEY,
    FCM_APP_ID
)

_LOGGER = logging.getLogger(__name__)

class Firebase():
    def __init__(self):
        self._pushClient = None

    async def firebaseStart(self, onNotificationFunc, firebaseCredentialsFileName: str, brand='cupra') -> bool: 
        """ Starts the firebase cloud messaging receiver """
        try:
            loop = asyncio.get_running_loop()
            credentials = await loop.run_in_executor(None, readFCMCredsFile, firebaseCredentialsFileName)
            #credentials = readFCMCredsFile(firebaseCredentialsFileName)

            #if credentials == {}:
            #    credentials =None

            fcm_project_id=FCM_PROJECT_ID
            fcm_app_id=FCM_APP_ID[brand]
            fcm_api_key=FCM_API_KEY
            chars = string.ascii_letters + string.digits
            fcmMessageSenderId = ''.join(secrets.choice(chars) for i in range(16)) 
            fcmMessageSenderId= 'fxpWQ_'+fcmMessageSenderId

            fcm_config = FcmRegisterConfig(fcm_project_id, fcm_app_id, fcm_api_key, fcmMessageSenderId)
            self._pushClient = FcmPushClient(onNotificationFunc, fcm_config, credentials, onFCMCredentialsUpdated)
            fcm_token = await self._pushClient.checkin_or_register(firebaseCredentialsFileName)
            _LOGGER.debug(f'Firebase.checkin_or_register() returned a token:{fcm_token}')
            await self._pushClient.start()
            await asyncio.sleep(5)
            return self._pushClient.is_started()
        except Exception as e:
            _LOGGER.error(f'Error in firebaseStart. Error: {e}')
            return False

    async def firebaseStop(self) -> bool: 
        """ Stops the firebase cloud messaging receiver """
        try:
            await self._pushClient.stop()
            #await asyncio.sleep(5)
            self._pushClient = None
            return True
        except Exception as e:
            _LOGGER.error(f'Error in firebaseStop. Error: {e}')
            return False

def readFCMCredsFile(credsFile) -> dict[str, Any]:
    """ Reads the firebase cloud messaging credentials from file"""
    try:
        if os.path.isfile(credsFile):
            with open(credsFile, "r") as f:
                credString=f.read()
            f.close()
            creds=json.loads(credString)
            return creds
        else:
            _LOGGER.debug(f'{credsFile} not found.')
            return {}
    except Exception as e:
        _LOGGER.warning(f'readFCMCredsFile() not successful. Error: {e}')
        return {}

def writeFCMCredsFile(creds, firebaseCredentialsFileName) -> None:
    """ Saves the firebase cloud messaging credentials to a file for future use """
    try:
        with open(firebaseCredentialsFileName, "w") as f:
            f.write(json.dumps(creds))
        f.close()
    except Exception as e:
        _LOGGER.warning(f'writeFCMCredsFile() not successful. Error: {e}')
    
async def onFCMCredentialsUpdated(creds: dict[str, Any], firebaseCredentialsFileName: str) -> None:
    """ Is called from firebase-messaging package """
    loop = asyncio.get_running_loop()
    await loop.run_in_executor(None, writeFCMCredsFile, creds, firebaseCredentialsFileName)
    #writeFCMCredsFile(creds, firebaseCredentialsFileName)
    



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2014 The Chromium Authors. All rights reserved.
Use of this source code is governed by a BSD-style license that can be
found in the LICENSE file.

Logging information for Android "checkin" events (automatic, periodic
requests made by Android devices to the server).
"""

import builtins
import google.protobuf.descriptor
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _DeviceType:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _DeviceTypeEnumTypeWrapper(
    google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_DeviceType.ValueType],
    builtins.type,
):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    DEVICE_ANDROID_OS: _DeviceType.ValueType  # 1
    """Android Device"""
    DEVICE_IOS_OS: _DeviceType.ValueType  # 2
    """Apple IOS device"""
    DEVICE_CHROME_BROWSER: _DeviceType.ValueType  # 3
    """Chrome browser - Not Chrome OS.  No hardware records."""
    DEVICE_CHROME_OS: _DeviceType.ValueType  # 4
    """Chrome OS"""

class DeviceType(_DeviceType, metaclass=_DeviceTypeEnumTypeWrapper):
    """enum values correspond to the type of device.
    Used in the AndroidCheckinProto and Device proto.
    """

DEVICE_ANDROID_OS: DeviceType.ValueType  # 1
"""Android Device"""
DEVICE_IOS_OS: DeviceType.ValueType  # 2
"""Apple IOS device"""
DEVICE_CHROME_BROWSER: DeviceType.ValueType  # 3
"""Chrome browser - Not Chrome OS.  No hardware records."""
DEVICE_CHROME_OS: DeviceType.ValueType  # 4
"""Chrome OS"""
global___DeviceType = DeviceType

@typing_extensions.final
class ChromeBuildProto(google.protobuf.message.Message):
    """Build characteristics unique to the Chrome browser, and Chrome OS"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _Platform:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _PlatformEnumTypeWrapper(
        google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[
            ChromeBuildProto._Platform.ValueType
        ],
        builtins.type,
    ):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        PLATFORM_WIN: ChromeBuildProto._Platform.ValueType  # 1
        PLATFORM_MAC: ChromeBuildProto._Platform.ValueType  # 2
        PLATFORM_LINUX: ChromeBuildProto._Platform.ValueType  # 3
        PLATFORM_CROS: ChromeBuildProto._Platform.ValueType  # 4
        PLATFORM_IOS: ChromeBuildProto._Platform.ValueType  # 5
        PLATFORM_ANDROID: ChromeBuildProto._Platform.ValueType  # 6
        """Just a placeholder. Likely don't need it due to the presence of the
        Android GCM on phone/tablet devices.
        """

    class Platform(_Platform, metaclass=_PlatformEnumTypeWrapper): ...
    PLATFORM_WIN: ChromeBuildProto.Platform.ValueType  # 1
    PLATFORM_MAC: ChromeBuildProto.Platform.ValueType  # 2
    PLATFORM_LINUX: ChromeBuildProto.Platform.ValueType  # 3
    PLATFORM_CROS: ChromeBuildProto.Platform.ValueType  # 4
    PLATFORM_IOS: ChromeBuildProto.Platform.ValueType  # 5
    PLATFORM_ANDROID: ChromeBuildProto.Platform.ValueType  # 6
    """Just a placeholder. Likely don't need it due to the presence of the
    Android GCM on phone/tablet devices.
    """

    class _Channel:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _ChannelEnumTypeWrapper(
        google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[
            ChromeBuildProto._Channel.ValueType
        ],
        builtins.type,
    ):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        CHANNEL_STABLE: ChromeBuildProto._Channel.ValueType  # 1
        CHANNEL_BETA: ChromeBuildProto._Channel.ValueType  # 2
        CHANNEL_DEV: ChromeBuildProto._Channel.ValueType  # 3
        CHANNEL_CANARY: ChromeBuildProto._Channel.ValueType  # 4
        CHANNEL_UNKNOWN: ChromeBuildProto._Channel.ValueType  # 5
        """for tip of tree or custom builds"""

    class Channel(_Channel, metaclass=_ChannelEnumTypeWrapper): ...
    CHANNEL_STABLE: ChromeBuildProto.Channel.ValueType  # 1
    CHANNEL_BETA: ChromeBuildProto.Channel.ValueType  # 2
    CHANNEL_DEV: ChromeBuildProto.Channel.ValueType  # 3
    CHANNEL_CANARY: ChromeBuildProto.Channel.ValueType  # 4
    CHANNEL_UNKNOWN: ChromeBuildProto.Channel.ValueType  # 5
    """for tip of tree or custom builds"""

    PLATFORM_FIELD_NUMBER: builtins.int
    CHROME_VERSION_FIELD_NUMBER: builtins.int
    CHANNEL_FIELD_NUMBER: builtins.int
    platform: global___ChromeBuildProto.Platform.ValueType
    """The platform of the device."""
    chrome_version: builtins.str
    """The Chrome instance's version."""
    channel: global___ChromeBuildProto.Channel.ValueType
    """The Channel (build type) of Chrome."""
    def __init__(
        self,
        *,
        platform: global___ChromeBuildProto.Platform.ValueType | None = ...,
        chrome_version: builtins.str | None = ...,
        channel: global___ChromeBuildProto.Channel.ValueType | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "channel",
            b"channel",
            "chrome_version",
            b"chrome_version",
            "platform",
            b"platform",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "channel",
            b"channel",
            "chrome_version",
            b"chrome_version",
            "platform",
            b"platform",
        ],
    ) -> None: ...

global___ChromeBuildProto = ChromeBuildProto

@typing_extensions.final
class AndroidCheckinProto(google.protobuf.message.Message):
    """Information sent by the device in a "checkin" request."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    LAST_CHECKIN_MSEC_FIELD_NUMBER: builtins.int
    CELL_OPERATOR_FIELD_NUMBER: builtins.int
    SIM_OPERATOR_FIELD_NUMBER: builtins.int
    ROAMING_FIELD_NUMBER: builtins.int
    USER_NUMBER_FIELD_NUMBER: builtins.int
    TYPE_FIELD_NUMBER: builtins.int
    CHROME_BUILD_FIELD_NUMBER: builtins.int
    last_checkin_msec: builtins.int
    """Miliseconds since the Unix epoch of the device's last successful checkin."""
    cell_operator: builtins.str
    """The current MCC+MNC of the mobile device's current cell."""
    sim_operator: builtins.str
    """The MCC+MNC of the SIM card (different from operator if the
    device is roaming, for instance).
    """
    roaming: builtins.str
    """The device's current roaming state (reported starting in eclair builds).
    Currently one of "{,not}mobile-{,not}roaming", if it is present at all.
    """
    user_number: builtins.int
    """For devices supporting multiple user profiles (which may be
    supported starting in jellybean), the ordinal number of the
    profile that is checking in.  This is 0 for the primary profile
    (which can't be changed without wiping the device), and 1,2,3,...
    for additional profiles (which can be added and deleted freely).
    """
    type: global___DeviceType.ValueType
    """Class of device.  Indicates the type of build proto
    (IosBuildProto/ChromeBuildProto/AndroidBuildProto)
    That is included in this proto
    """
    @property
    def chrome_build(self) -> global___ChromeBuildProto:
        """For devices running MCS on Chrome, build-specific characteristics
        of the browser.  There are no hardware aspects (except for ChromeOS).
        This will only be populated for Chrome builds/ChromeOS devices
        """
    def __init__(
        self,
        *,
        last_checkin_msec: builtins.int | None = ...,
        cell_operator: builtins.str | None = ...,
        sim_operator: builtins.str | None = ...,
        roaming: builtins.str | None = ...,
        user_number: builtins.int | None = ...,
        type: global___DeviceType.ValueType | None = ...,
        chrome_build: global___ChromeBuildProto | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing_extensions.Literal[
            "cell_operator",
            b"cell_operator",
            "chrome_build",
            b"chrome_build",
            "last_checkin_msec",
            b"last_checkin_msec",
            "roaming",
            b"roaming",
            "sim_operator",
            b"sim_operator",
            "type",
            b"type",
            "user_number",
            b"user_number",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing_extensions.Literal[
            "cell_operator",
            b"cell_operator",
            "chrome_build",
            b"chrome_build",
            "last_checkin_msec",
            b"last_checkin_msec",
            "roaming",
            b"roaming",
            "sim_operator",
            b"sim_operator",
            "type",
            b"type",
            "user_number",
            b"user_number",
        ],
    ) -> None: ...

global___AndroidCheckinProto = AndroidCheckinProto



"""Constants for pycupra library."""

BASE_SESSION = 'https://ola.prod.code.seat.cloud.vwgroup.com'
BASE_AUTH = 'https://identity.vwgroup.io'

# Data used in communication
CLIENT_LIST = {
    'seat': {
        'CLIENT_ID': '99a5b77d-bd88-4d53-b4e5-a539c60694a3@apps_vw-dilab_com',
        'SCOPE': 'openid profile nickname birthdate phone',
        'REDIRECT_URL': 'seat://oauth-callback',
        'TOKEN_TYPES': 'code id_token token'
    },
    'cupra': {
        'CLIENT_ID': '3c756d46-f1ba-4d78-9f9a-cff0d5292d51@apps_vw-dilab_com',
        'CLIENT_SECRET': 'eb8814e641c81a2640ad62eeccec11c98effc9bccd4269ab7af338b50a94b3a2',
        'SCOPE': 'openid profile nickname birthdate phone',
        'REDIRECT_URL': 'cupra://oauth-callback',
        'TOKEN_TYPES': 'code id_token token'
    }
}


XCLIENT_ID = '3c756d46-f1ba-4d78-9f9a-cff0d5292d51@apps_vw-dilab.com' 
XAPPVERSION = '2.5.0'
XAPPNAME = 'com.cupra.mycupra'
USER_AGENT_CUPRA = 'OLACupra/2.5.0 (Android 12; sdk_gphone64_x86_64; Google) Mobile' 
USER_AGENT_SEAT = 'OLASeat/2.5.0 (Android 12; sdk_gphone64_x86_64; Google) Mobile'
APP_URI = 'https://ola.prod.code.seat.cloud.vwgroup.com'

HEADERS_SESSION = {
    'seat': {
    'Connection': 'keep-alive',
    'Content-Type': 'application/json',
    'Accept-charset': 'UTF-8',
    'Accept': 'application/json',
    #'X-Client-Id': XCLIENT_ID,
    #'X-App-Version': XAPPVERSION,
    #'X-App-Name': XAPPNAME,
    'User-Agent': USER_AGENT_SEAT
    #'User-ID': '?????', # to be set later
    },
    'cupra': {
    'Connection': 'keep-alive',
    'Content-Type': 'application/json',
    'Accept-charset': 'UTF-8',
    'Accept': 'application/json',
    #'X-Client-Id': XCLIENT_ID,
    #'X-App-Version': XAPPVERSION,
    #'X-App-Name': XAPPNAME,
    'User-Agent': USER_AGENT_CUPRA
    #'User-ID': '?????', # to be set later
    }
}

HEADERS_AUTH = {
    'seat': {
        'Content-Type': 'application/x-www-form-urlencoded',
        'User-Agent': USER_AGENT_SEAT,
        'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9',
        'x-requested-with': XAPPNAME,
        'Accept-Encoding': 'gzip, deflate',
        'Connection': 'keep-alive',
        #'X-App-Name': XAPPNAME
    },
    'cupra':{
        'Content-Type': 'application/x-www-form-urlencoded',
        'User-Agent': USER_AGENT_CUPRA,
        'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9',
        'x-requested-with': XAPPNAME,
        'Accept-Encoding': 'gzip, deflate',
        'Connection': 'keep-alive',
        #'X-App-Name': XAPPNAME
    }
}

TOKEN_HEADERS = {
    'seat': {
        'Accept': 'application/json',
        'X-Platform': 'Android',
        #'X-Language-Id': 'XX',
        #'X-Country-Id': 'XX',
        #'Accept-Language': 'XX',
        'Accept-Charset': 'UTF-8',
        'Content-Type': 'application/x-www-form-urlencoded',
        'Connection': 'keep-alive',
        'Accept-Encoding': 'gzip',
        'User-Agent': USER_AGENT_SEAT,
        'app-version': '2.5.0',
        'app-brand': 'seat',
        'app-market': 'android',
        #'User-ID': '?????', # to be set later
        'Authorization': 'Bearer'
    },
    'cupra': {
        'Accept': 'application/json',
        'X-Platform': 'Android',
        #'X-Language-Id': 'XX',
        #'X-Country-Id': 'XX',
        #'Accept-Language': 'XX',
        'Accept-Charset': 'UTF-8',
        'Content-Type': 'application/x-www-form-urlencoded',
        'Connection': 'keep-alive',
        'Accept-Encoding': 'gzip',
        'User-Agent': USER_AGENT_CUPRA,
        'app-version': '2.5.0',
        'app-brand': 'cupra',
        'app-market': 'android',
        #'User-ID': '?????', # to be set later
        'Authorization': 'Bearer'
    }
}

#ERROR_CODES = {
#    '11': 'Charger not connected'
#}

### API Endpoints below, not yet in use ###
# API AUTH endpoints
AUTH_OIDCONFIG = 'https://identity.vwgroup.io/.well-known/openid-configuration'                     # OpenID configuration
AUTH_TOKEN = 'https://identity.vwgroup.io/oidc/v1/token'                                      # Endpoint for exchanging code for token
AUTH_REFRESH = 'https://ola.prod.code.seat.cloud.vwgroup.com/authorization/api/v1/token'             # Endpoint for token refresh (also used for exchanging code for token for Seat)
AUTH_TOKENKEYS = 'https://identity.vwgroup.io/oidc/v1/keys'                                         # Signing keys for tokens

# API endpoints
API_MBB_STATUSDATA = 'https://customer-profile.vwgroup.io/v3/customers/{self._user_id}/mbbStatusData'
API_PERSONAL_DATA= 'https://customer-profile.vwgroup.io/v3/customers/{self._user_id}/personalData'
#Other option for personal data is '{baseurl}/v1/users/{self._user_id}'

API_VEHICLES = '{APP_URI}/v2/users/{self._user_id}/garage/vehicles'                                  # Garage info
API_MYCAR = '{baseurl}/v5/users/{self._user_id}/vehicles/{vin}/mycar'                                # Vehicle status report
API_CHARGING = '{baseurl}/v1/vehicles/{vin}/charging'                                                # Vehicle charging information 
#API_OPERLIST = '{homeregion}/api/rolesrights/operationlist/v3/vehicles/{vin}'                       # API Endpoint for supported operations
#API_CHARGER = 'fs-car/bs/batterycharge/v1/{BRAND}/{COUNTRY}/vehicles/{vin}/charger'                 # Charger data
API_CLIMATER_STATUS = '{baseurl}/v1/vehicles/{vin}/climatisation/status'                             # Climatisation data
API_CLIMATER = '{baseurl}/v2/vehicles/{vin}/climatisation'                         # Climatisation data
API_DEPARTURE_TIMERS = '{baseurl}/v1/vehicles/{vin}/departure-timers'                                 # Departure timers
API_DEPARTURE_PROFILES = '{baseurl}/v1/vehicles/{vin}/departure/profiles'                             # Departure profiles
API_POSITION = '{baseurl}/v1/vehicles/{vin}/parkingposition'                                         # Position data
API_POS_TO_ADDRESS= 'https://maps.googleapis.com/maps/api/directions/json?origin={lat},{lon}&destination={lat},{lon}&traffic_model=best_guess&departure_time=now&language=de&key={apiKeyForGoogle}&mode=driving'
API_TRIP = '{baseurl}/v1/vehicles/{vin}/driving-data/{dataType}?from={startDate}T00:00:00Z&to=2099-12-31T09:59:01Z' # Trip statistics (whole history) SHORT/LONG/CYCLIC (WEEK only with from)
API_MILEAGE = '{baseurl}/v1/vehicles/{vin}/mileage'                                                  # Total km etc
API_MAINTENANCE = '{baseurl}/v1/vehicles/{vin}/maintenance'                                          # Inspection information
API_MEASUREMENTS = '{baseurl}/v1/vehicles/{vin}/measurements/engines'                                # ???
API_STATUS = '{baseurl}/v2/vehicles/{vin}/status'                                                    # Status information like locks and windows
API_WARNINGLIGHTS = '{baseurl}/v3/vehicles/{vin}/warninglights'                                      # ???
API_SHOP = '{baseurl}/v1/shop/vehicles/{vin}/articles'                                               # ???
#API_ACTION = '{baseurl}/v1/vehicles/{vin}/{action}/requests/{command}'                               # Actions (e.g. ActionCharge="charging", ActionChargeStart="start",ActionChargeStop="stop")
API_RELATION_STATUS = '{baseurl}/v1/users/{self._user_id}/vehicles/{vin}/relation-status'            # ???
API_INVITATIONS = '{baseurl}/v1/user/{self._user_id}/invitations'                                    # ???
API_CAPABILITIES = '{APP_URI}/v1/user/{self._user_id}/vehicle/{vin}/capabilities'                    # ???
#API_CAPABILITIES_MANAGEMENT = '{API_CAPABILITIES}/management'                                        # ???
API_IMAGE = '{baseurl}/v1/vehicles/{vin}/renders'
API_HONK_AND_FLASH = '{baseurl}//v1/vehicles/{vin}/honk-and-flash'
API_ACCESS = '{baseurl}//v1/vehicles/{vin}/access/{action}'                                          # to lock or unlock vehicle
API_REQUESTS = '{baseurl}/vehicles/{vin}/{capability}/requests'
API_REFRESH = '{baseurl}/v1/vehicles/{vin}/vehicle-wakeup/request'
API_SECTOKEN = '{baseurl}/v2/users/{self._user_id}/spin/verify'
API_DESTINATION = '{baseurl}/v1/users/vehicles/{vin}/destination'
API_LITERALS= '{APP_URI}/v1/content/apps/my-cupra/literals/{language}'                               # Message texts in different langauages, e.g. 'en_GB'

# Still to analyse if needed
#'{baseurl}/settings/api/v1?vin={vin}&vehicle-model=LeonST&region=US&enrolment-country=DE&platform=MOD3'
#'{baseurl}/v1/users/{self._user_id}/vin/{vin}/terms-and-conditions'
#'{baseurl}/v2/subscriptions'
#'{baseurl}/v1/users/{self._user_id}/vehicles/{vin}/leads/history'
#'{baseurl}/v1/users/{self._user-id}/vehicles/{vin}/consents/xxcryptickeyxxx?locale=en_DE' #{"userId":"xxxxxxx","locale":"en_DE","error":{"title":"Consent failed to load.","detail":"CUPRAApp_ME3_Vehicle_VehiclePermissions_MainViewAccepted_Low_Type1_Wrong"}}

API_CONNECTION= '{APP_URI}/vehicles/{vin}/connection'
#API_CONSENTS='{APP_URI}/v1/users/{self._user_id}/consents'
API_PSP='{baseurl}/v2/users/{self._user_id}/vehicles/{vin}/psp'     # primary service provider (Werkstatt)
API_USER_INFO= 'https://identity-userinfo.vwgroup.io/oidc/userinfo' #{"sub":"xxx","name":"xxx","given_name":"xxx","family_name":"xxx","nickname":"xxx","email":"###","email_verified":true,"birthdate":"###","updated_at":123456789,"picture":"https://customer-pictures.vwgroup.io/v1/###/profile-picture"}

PUBLIC_MODEL_IMAGES_SERVER = 'prod-ola-public-bucket.s3.eu-central-1.amazonaws.com'                      # non-indivdual model images are on this server

# API endpoints for status
REQ_STATUS = {
    'climatisation': 'fs-car/bs/climatisation/v1/{BRAND}/{COUNTRY}/vehicles/{vin}/climater/actions/{id}',
    'batterycharge': 'fs-car/bs/batterycharge/v1/{BRAND}/{COUNTRY}/vehicles/{vin}/charger/actions/{id}',
    'departuretimer': 'fs-car/bs/departuretimer/v1/{BRAND}/{COUNTRY}/vehicles/{vin}/timer/actions/{id}',
    'vsr': 'fs-car/bs/vsr/v1/{BRAND}/{COUNTRY}/vehicles/{vin}/requests/{id}/jobstatus',
    'default': 'fs-car/bs/{section}/v1/{BRAND}/{COUNTRY}/vehicles/{vin}/requests/{id}/status'
}

FCM_PROJECT_ID='ola-apps-prod'
FCM_APP_ID={
    'cupra': '1:530284123617:android:9b9ba5a87c7ffd37fbeea0',
    'seat':  '1:530284123617:android:d6187613ac3d7b08fbeea0'
}
FCM_API_KEY='AIzaSyCoSp1zitklb1EDj5yQumN0VNhDizJQHLk'
FIREBASE_STATUS_NOT_INITIALISED= 0
FIREBASE_STATUS_ACTIVATED= 1
FIREBASE_STATUS_NOT_WANTED= -2
FIREBASE_STATUS_ACTIVATION_FAILED= -1
FIREBASE_STATUS_ACTIVATION_STOPPED= -3



from datetime import date, datetime
from base64 import b64encode
from string import ascii_letters as letters, digits
from sys import argv
from os import environ as env
from os.path import join, dirname, expanduser
from itertools import product
import json
import logging
import re

_LOGGER = logging.getLogger(__name__)


def json_loads(s):
    return json.loads(s, object_hook=obj_parser)


def obj_parser(obj):
    """Parse datetime."""
    for key, val in obj.items():
        try:
            obj[key] = datetime.strptime(val, "%Y-%m-%dT%H:%M:%S%z")
        except (TypeError, ValueError):
            pass
    return obj


def find_path(src, path):
    """Simple navigation of a hierarchical dict structure using XPATH-like syntax.

    >>> find_path(dict(a=1), 'a')
    1

    >>> find_path(dict(a=1), '')
    {'a': 1}

    >>> find_path(dict(a=None), 'a')


    >>> find_path(dict(a=1), 'b')
    Traceback (most recent call last):
    ...
    KeyError: 'b'

    >>> find_path(dict(a=dict(b=1)), 'a.b')
    1

    >>> find_path(dict(a=dict(b=1)), 'a')
    {'b': 1}

    >>> find_path(dict(a=dict(b=1)), 'a.c')
    Traceback (most recent call last):
    ...
    KeyError: 'c'

    """
    if not path:
        return src
    if isinstance(path, str):
        path = path.split(".")
    return find_path(src[path[0]], path[1:])


def is_valid_path(src, path) -> bool:
    """
    >>> is_valid_path(dict(a=1), 'a')
    True

    >>> is_valid_path(dict(a=1), '')
    True

    >>> is_valid_path(dict(a=1), None)
    True

    >>> is_valid_path(dict(a=1), 'b')
    False
    """
    try:
        find_path(src, path)
        return True
    except KeyError:
        return False


def camel2slug(s) -> str:
    """Convert camelCase to camel_case.

    >>> camel2slug('fooBar')
    'foo_bar'
    """
    return re.sub("([A-Z])", "_\\1", s).lower().lstrip("_")



from .fcmpushclient import FcmPushClient, FcmPushClientConfig, FcmPushClientRunState
from .fcmregister import FcmRegisterConfig
from .mcs_pb2 import (  # pylint: disable=no-name-in-module
    Close,
    DataMessageStanza,
    HeartbeatAck,
    HeartbeatPing,
    IqStanza,
    LoginRequest,
    LoginResponse,
    SelectiveAck,
    StreamErrorStanza,
)

from .android_checkin_pb2 import (
    DEVICE_CHROME_BROWSER,
    AndroidCheckinProto,
    ChromeBuildProto,
)
from .checkin_pb2 import (
    AndroidCheckinRequest,
    AndroidCheckinResponse,
)

__all__ = [
    "FcmPushClientConfig",
    "FcmPushClient",
    "FcmPushClientRunState",
    "FcmRegisterConfig",
]



class SeatConfigException(Exception):
    """Raised when Seat Connect API client is configured incorrectly"""

    def __init__(self, status):
        """Initialize exception"""
        super(SeatConfigException, self).__init__(status)
        self.status = status

class SeatAuthenticationException(Exception):
    """Raised when credentials are invalid during authentication"""

    def __init__(self, status):
        """Initialize exception"""
        super(SeatAuthenticationException, self).__init__(status)
        self.status = status

class SeatAccountLockedException(Exception):
    """Raised when account is locked from too many login attempts"""

    def __init__(self, status):
        """Initialize exception"""
        super(SeatAccountLockedException, self).__init__(status)
        self.status = status

class SeatTokenExpiredException(Exception):
    """Raised when server reports that the access token has expired"""

    def __init__(self, status):
        """Initialize exception"""
        super(SeatTokenExpiredException, self).__init__(status)
        self.status = status

class SeatException(Exception):
    """Raised when an unknown error occurs during API interaction"""

    def __init__(self, status):
        """Initialize exception"""
        super(SeatException, self).__init__(status)
        self.status = status

class SeatThrottledException(Exception):
    """Raised when the API throttles the connection"""

    def __init__(self, status):
        """Initialize exception"""
        super(SeatThrottledException, self).__init__(status)
        self.status = status

class SeatEULAException(Exception):
    """Raised when EULA must be accepted before login"""

    def __init__(self, status):
        """Initialize exception"""
        super(SeatEULAException, self).__init__(status)
        self.status = status

class SeatLoginFailedException(Exception):
    """Raised when login fails for an unknown reason"""

    def __init__(self, status):
        """Initialize exception"""
        super(SeatLoginFailedException, self).__init__(status)
        self.status = status

class SeatInvalidRequestException(Exception):
    """Raised when an unsupported request is made"""

    def __init__(self, status):
        """Initialize exception"""
        super(SeatInvalidRequestException, self).__init__(status)
        self.status = status

class SeatRequestInProgressException(Exception):
    """Raised when a request fails because another request is already in progress"""

    def __init__(self, status):
        """Initialize exception"""
        super(SeatRequestInProgressException, self).__init__(status)
        self.status = status

class SeatServiceUnavailable(Exception):
    """Raised when a API is unavailable"""

    def __init__(self, status):
        """Initialize exception"""
        super(SeatServiceUnavailable, self).__init__(status)
        self.status = status



"""Constants module."""

GCM_REGISTER_URL = "https://android.clients.google.com/c2dm/register3"
GCM_CHECKIN_URL = "https://android.clients.google.com/checkin"
GCM_SERVER_KEY_BIN = (
    b"\x04\x33\x94\xf7\xdf\xa1\xeb\xb1\xdc\x03\xa2\x5e\x15\x71\xdb\x48\xd3"
    + b"\x2e\xed\xed\xb2\x34\xdb\xb7\x47\x3a\x0c\x8f\xc4\xcc\xe1\x6f\x3c"
    + b"\x8c\x84\xdf\xab\xb6\x66\x3e\xf2\x0c\xd4\x8b\xfe\xe3\xf9\x76\x2f"
    + b"\x14\x1c\x63\x08\x6a\x6f\x2d\xb1\x1a\x95\xb0\xce\x37\xc0\x9c\x6e"
)
# urlsafe b64 encoding of the binary key with = padding removed
GCM_SERVER_KEY_B64 = (
    "BDOU99-h67HcA6JeFXHbSNMu7e2yNNu3RzoM"
    + "j8TM4W88jITfq7ZmPvIM1Iv-4_l2LxQcYwhqby2xGpWwzjfAnG4"
)

FCM_SUBSCRIBE_URL = "https://fcm.googleapis.com/fcm/connect/subscribe/"
FCM_SEND_URL = "https://fcm.googleapis.com/fcm/send/"

FCM_API = "https://fcm.googleapis.com/v1/"
FCM_REGISTRATION = "https://fcmregistrations.googleapis.com/v1/"
FCM_INSTALLATION = "https://firebaseinstallations.googleapis.com/v1/"
AUTH_VERSION = "FIS_v2"
SDK_VERSION = "w:0.6.6"

DOORBELLS_ENDPOINT = "/clients_api/doorbots/{0}"

MCS_VERSION = 41
MCS_HOST = "mtalk.google.com"
MCS_PORT = 5228
MCS_SELECTIVE_ACK_ID = 12
MCS_STREAM_ACK_ID = 13



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Communicate with the My Cupra portal."""
"""First fork from https://github.com/robinostlund/volkswagencarnet where it was modified to support also Skoda Connect"""
"""Then forked from https://github.com/lendy007/skodaconnect for adaptation to Seat Connect"""
"""Then forked from https://github.com/Farfar/seatconnect for adaptation to the new API of My Cupra and My Seat"""
import re
import os
import json
import logging
import asyncio
import hashlib
import jwt
import string
import secrets
import xmltodict
from copy import deepcopy
import importlib.metadata
from typing import Any

from PIL import Image
from io import BytesIO
from sys import version_info, argv
from datetime import timedelta, datetime, timezone
from urllib.parse import urljoin, parse_qs, urlparse, urlencode
from json import dumps as to_json
from jwt.exceptions import ExpiredSignatureError
import aiohttp
from bs4 import BeautifulSoup
from base64 import b64decode, b64encode, urlsafe_b64decode, urlsafe_b64encode
#from .__version__ import __version__ as lib_version
from .utilities import json_loads
from .vehicle import Vehicle
from .exceptions import (
    SeatConfigException,
    SeatAuthenticationException,
    SeatAccountLockedException,
    SeatTokenExpiredException,
    SeatException,
    SeatEULAException,
    SeatThrottledException,
    SeatLoginFailedException,
    SeatInvalidRequestException,
    SeatRequestInProgressException,
    SeatServiceUnavailable
)

from requests_oauthlib import OAuth2Session
from oauthlib.oauth2.rfc6749.parameters import parse_authorization_code_response, parse_token_response, prepare_grant_uri

from aiohttp import ClientSession, ClientTimeout
from aiohttp.hdrs import METH_GET, METH_POST, METH_PUT, METH_DELETE

from .const import (
    HEADERS_SESSION,
    HEADERS_AUTH,
    TOKEN_HEADERS,
    BASE_SESSION,
    BASE_AUTH,
    CLIENT_LIST,
    XCLIENT_ID,
    XAPPVERSION,
    XAPPNAME,
    AUTH_OIDCONFIG,
    AUTH_TOKEN,
    AUTH_TOKENKEYS,
    AUTH_REFRESH,
    APP_URI,
    API_MBB_STATUSDATA,
    API_PERSONAL_DATA,
    API_USER_INFO,
    API_CONNECTION,
    API_PSP,
    API_VEHICLES,
    API_MYCAR,
    API_STATUS,
    API_CHARGING,
    API_POSITION,
    API_POS_TO_ADDRESS,
    API_TRIP,
    API_CLIMATER_STATUS,
    API_CLIMATER,
    API_DEPARTURE_TIMERS,
    API_DEPARTURE_PROFILES,
    API_MILEAGE,
    API_CAPABILITIES,
    #API_CAPABILITIES_MANAGEMENT,
    API_MAINTENANCE,
    API_WARNINGLIGHTS,
    API_MEASUREMENTS,
    API_RELATION_STATUS,
    API_INVITATIONS,
    #API_ACTION,
    API_IMAGE,
    API_HONK_AND_FLASH,
    API_ACCESS,
    API_SECTOKEN,
    API_REQUESTS,
    API_REFRESH,
    API_DESTINATION,

    PUBLIC_MODEL_IMAGES_SERVER,
    FIREBASE_STATUS_NOT_INITIALISED,
)

version_info >= (3, 0) or exit('Python 3 required')
lib_version = importlib.metadata.version("pycupra")

_LOGGER = logging.getLogger(__name__)
BRAND_CUPRA = 'cupra'
TIMEOUT = timedelta(seconds=90)

class Connection:
    """ Connection to Connect services """
  # Init connection class
    def __init__(self, session, brand='cupra', username='', password='', fulldebug=False, nightlyUpdateReduction=False, anonymise=True, tripStatisticsStartDate=None, **optional):
        """ Initialize """
        self._session = session
        self._lock = asyncio.Lock()
        self._session_fulldebug = fulldebug
        self._session_nightlyUpdateReduction = nightlyUpdateReduction
        self._session_anonymise = anonymise
        self._session_tripStatisticsStartDate = tripStatisticsStartDate
        self._session_headers = HEADERS_SESSION.get(brand).copy()
        self._session_base = BASE_SESSION
        self._session_auth_headers = HEADERS_AUTH.copy()
        self._session_token_headers = TOKEN_HEADERS.copy()
        self._session_cookies = ""
        self._session_nonce = self._getNonce()
        self._session_state = self._getState()

        self._session_auth_ref_url = BASE_SESSION
        self._session_spin_ref_url = BASE_SESSION
        self._session_first_update = False
        self._session_auth_brand = brand
        self._session_auth_username = username
        self._session_auth_password = password
        self._session_tokens = {}

        self._vehicles = []
        self._userData = {}

        _LOGGER.info(f'Init PyCupra library, version {lib_version}')
        _LOGGER.debug(f'Using service {self._session_base}')

        self._sessionRequestCounter = 0
        self._sessionRequestTimestamp = datetime.now(tz= None)
        self._sessionRequestCounterHistory = {}
        self._anonymisationDict={}
        self.addToAnonymisationDict(self._session_auth_username, '[USERNAME_ANONYMISED]')
        self.addToAnonymisationDict(self._session_auth_password, '[PASSWORD_ANONYMISED]')
        self._anonymisationKeys={'firstName', 'lastName', 'dateOfBirth', 'nickname'}
        self.addToAnonymisationKeys('name')
        self.addToAnonymisationKeys('given_name')
        self.addToAnonymisationKeys('email')
        self.addToAnonymisationKeys('family_name')
        self.addToAnonymisationKeys('birthdate')
        self.addToAnonymisationKeys('vin')
        self._error401 = False


    def _clear_cookies(self):
        self._session._cookie_jar._cookies.clear()
        self._session_cookies = ''

    def _getNonce(self):
        chars = string.ascii_letters + string.digits
        text = ''.join(secrets.choice(chars) for i in range(10))
        sha256 = hashlib.sha256()
        sha256.update(text.encode())
        return (b64encode(sha256.digest()).decode('utf-8')[:-1])

    def _getState(self):
        return self._getNonce()

    def readTokenFile(self, brand):
        try:
            if os.path.isfile(self._tokenFile):
                with open(self._tokenFile, "r") as f:
                    tokenString=f.read()
                f.close()
                tokens=json.loads(tokenString)
                self._session_tokens[brand]=tokens
                self._user_id=tokens['user_id']
                self.addToAnonymisationDict(self._user_id,'[USER_ID_ANONYMISED]')
                return True
            _LOGGER.info('No token file present. readTokenFile() returns False.')
            return False
        except:
            _LOGGER.warning('readTokenFile() not successful.')
            return False

    def writeTokenFile(self, brand):
        if hasattr(self, '_tokenfile'):
            _LOGGER.info('No token file name provided. Cannot write tokens to file.')
            return False
        self._session_tokens[brand]['user_id']=self._user_id
        try:
            with open(self._tokenFile, "w") as f:
                f.write(json.dumps(self._session_tokens[brand]))
            f.close()
            return True
        except Exception as e:
            _LOGGER.warning(f'writeTokenFile() not successful. Error: {e}')
            return False

    def deleteTokenFile(self):
        if hasattr(self, '_tokenfile'):
            _LOGGER.debug('No token file name provided. Cannot delete token file.')
            return False
        try:
            os.remove(self._tokenFile)
            _LOGGER.info(f'Deleted token file.')
            return True
        except Exception as e:
            _LOGGER.warning(f'deleteTokenFile() not successful. Error: {e}')
            return False

    def writeImageFile(self, imageName, imageData, imageDict):
        try:
            with open(f'./www/image_{imageName}.png', "wb") as f:
                f.write(imageData)
            imageDict[imageName]=f'/local/image_{imageName}.png'
            f.close()
            return True
        except:
            _LOGGER.warning('writeImageFile() not successful. Ignoring this problem.')
            return False

  # API login/logout/authorization
    async def doLogin(self,**data) -> bool:
        """Login method, clean login or use token from file and refresh it"""
        #if len(self._session_tokens) > 0:
        #    _LOGGER.info('Revoking old tokens.')
        #    try:
        #        await self.logout()
        #    except:
        #        pass
        _LOGGER.info('doLogin() first tries to read tokens from file and to refresh them.')

        # Remove cookies and re-init session
        self._clear_cookies()
        self._vehicles.clear()
        self._session_tokens = {}
        self._session_headers = HEADERS_SESSION.get(self._session_auth_brand, HEADERS_SESSION['cupra']).copy()
        self._session_auth_headers = HEADERS_AUTH.copy()
        self._session_nonce = self._getNonce()
        self._session_state = self._getState()

        if data.get('apiKey',None)!=None:
            self._googleApiKey=data.get('apiKey')
        if data.get('tokenFile',None)!=None:
            self._tokenFile=data.get('tokenFile')
            loop = asyncio.get_running_loop()
            result = await loop.run_in_executor(None, self.readTokenFile, self._session_auth_brand)
            if result:
                rc=await self.refresh_token(self._session_auth_brand)
                if rc:
                    _LOGGER.info('Successfully read tokens from file and refreshed them.')
                    return True
        _LOGGER.info('Initiating new login with user name and password.')
        return await self._authorize(self._session_auth_brand)

    async def _authorize(self, client=BRAND_CUPRA) -> bool:
        """"Login" function. Authorize a certain client type and get tokens."""
        # Helper functions
        def extract_csrf(req):
            return re.compile('<meta name="_csrf" content="([^"]*)"/>').search(req).group(1)

        def extract_guest_language_id(req):
            return req.split('_')[1].lower()

        # Login/Authorization starts here
        try:
            #self._session_headers = HEADERS_SESSION.get(client).copy()
            #self._session_auth_headers = HEADERS_AUTH.copy()

            _LOGGER.debug(f'Starting authorization process for client {client}')
            req = await self._session.get(
                url=AUTH_OIDCONFIG
            )
            if req.status != 200:
                _LOGGER.debug(f'Get request to {AUTH_OIDCONFIG} was not successful. Response: {req}')
                return False
            response_data =  await req.json()
            authorizationEndpoint = response_data['authorization_endpoint']
            authissuer = response_data['issuer']
            oauthClient = OAuth2Session(client_id=CLIENT_LIST[client].get('CLIENT_ID'), scope=CLIENT_LIST[client].get('SCOPE'), redirect_uri=CLIENT_LIST[client].get('REDIRECT_URL'))
            code_verifier = urlsafe_b64encode(os.urandom(40)).decode('utf-8')
            code_verifier = re.sub('[^a-zA-Z0-9]+', '', code_verifier)
            code_challenge_hash = hashlib.sha256(code_verifier.encode("utf-8")).digest()
            code_challenge = urlsafe_b64encode(code_challenge_hash).decode("utf-8")
            code_challenge = code_challenge.replace("=", "")
            authorization_url, state = oauthClient.authorization_url(authorizationEndpoint, code_challenge=code_challenge, code_challenge_method='S256', 
                                                                        nonce=self._session_nonce, state=self._session_state)

            # Get authorization page (login page)
            if self._session_fulldebug:
                _LOGGER.debug(f'Get authorization page: "{authorization_url}"')
            try:
                req = await self._session.get(
                    url=authorization_url,
                        headers=self._session_auth_headers.get(client),
                        allow_redirects=False
                    )
                if req.headers.get('Location', False):
                    ref = req.headers.get('Location', '')
                    if 'error' in ref:
                        error = parse_qs(urlparse(ref).query).get('error', '')[0]
                        if 'error_description' in ref:
                            error = parse_qs(urlparse(ref).query).get('error_description', '')[0]
                            _LOGGER.info(f'Unable to login, {error}')
                        else:
                            _LOGGER.info(f'Unable to login.')
                        raise SeatException(error)
                    else:
                        if self._session_fulldebug:
                            _LOGGER.debug(f'Got authorization endpoint: "{ref}"')
                        req = await self._session.get(
                            url=ref,
                            headers=self._session_auth_headers.get(client),
                            allow_redirects=False
                        )
                else:
                    _LOGGER.warning(f'Unable to fetch authorization endpoint')
                    raise SeatException('Missing "location" header')
            except (SeatException):
                raise
            except Exception as error:
                _LOGGER.warning(f'Failed to get authorization endpoint. {error}')
                raise SeatException(error)

            # If we need to sign in (first token)
            if 'signin-service' in ref:
                _LOGGER.debug("Got redirect to signin-service")
                location = await self._signin_service(req, authissuer, authorizationEndpoint, client)
            else:
                # We are already logged on, shorter authorization flow
                location = req.headers.get('Location', None)

            # Follow all redirects until we reach the callback URL
            try:
                maxDepth = 10
                while not location.startswith(CLIENT_LIST[client].get('REDIRECT_URL')):
                    if location is None:
                        raise SeatException('Login failed')
                    if 'error' in location:
                        errorTxt = parse_qs(urlparse(location).query).get('error', '')[0]
                        if errorTxt == 'login.error.throttled':
                            timeout = parse_qs(urlparse(location).query).get('enableNextButtonAfterSeconds', '')[0]
                            raise SeatAccountLockedException(f'Account is locked for another {timeout} seconds')
                        elif errorTxt == 'login.errors.password_invalid':
                            raise SeatAuthenticationException('Invalid credentials')
                        else:
                            _LOGGER.warning(f'Login failed: {errorTxt}')
                        raise SeatLoginFailedException(errorTxt)
                    if 'terms-and-conditions' in location:
                        raise SeatEULAException('The terms and conditions must be accepted first at your local SEAT/Cupra site, e.g. "https://cupraofficial.se/"')
                    if 'user_id' in location: # Get the user_id which is needed for some later requests
                        self._user_id=parse_qs(urlparse(location).query).get('user_id', [''])[0]
                        self.addToAnonymisationDict(self._user_id,'[USER_ID_ANONYMISED]')
                        #_LOGGER.debug('Got user_id: %s' % self._user_id)
                    if self._session_fulldebug:
                        _LOGGER.debug(self.anonymise(f'Following redirect to "{location}"'))
                    response = await self._session.get(
                        url=location,
                        headers=self._session_auth_headers.get(client),
                        allow_redirects=False
                    )
                    if response.headers.get('Location', False) is False:
                        _LOGGER.debug(f'Unexpected response: {await req.text()}')
                        raise SeatAuthenticationException('User appears unauthorized')
                    location = response.headers.get('Location', None)
                    # Set a max limit on requests to prevent forever loop
                    maxDepth -= 1
                    if maxDepth == 0:
                        raise SeatException('Too many redirects')
            except (SeatException, SeatEULAException, SeatAuthenticationException, SeatAccountLockedException, SeatLoginFailedException):
                raise
            except Exception as e:
                # If we get an unhandled exception it should be because we can't redirect to the APP_URI URL and thus we have our auth code
                if 'code' in location:
                    if self._session_fulldebug:
                        _LOGGER.debug('Got code: %s' % location)
                    pass
                else:
                    _LOGGER.debug(f'Exception occured while logging in.')
                    raise SeatLoginFailedException(e)

            _LOGGER.debug('Received authorization code, exchange for tokens.')
            # Extract code and tokens
            auth_code = parse_qs(urlparse(location).query).get('code', [''])[0]
            # Save access, identity and refresh tokens according to requested client"""
            if client=='cupra':
                # oauthClient.fetch_token() does not work in home assistant, using POST request instead
                #token_data= oauthClient.fetch_token(token_url=AUTH_TOKEN,  
                #                       client_id=CLIENT_LIST[client].get('CLIENT_ID'), client_secret=CLIENT_LIST[client].get('CLIENT_SECRET'), authorization_response=location, 
                #                       code_verifier=code_verifier, code=auth_code)
                data = {
                        'redirect_uri': CLIENT_LIST[client].get('REDIRECT_URL'),
                        'client_id': CLIENT_LIST[client].get('CLIENT_ID'),
                        'client_secret': CLIENT_LIST[client].get('CLIENT_SECRET'), 
                        'authorization_response': location,
                        'code': auth_code,
                        'code_verifier': code_verifier,
                        'grant_type': 'authorization_code'
                        }
                req = await self._session.post(
                    url=AUTH_TOKEN,
                    data = data,
                    headers=self._session_auth_headers.get(client),
                    allow_redirects=False
                    )
                token_data = await req.json()
            else:
                data = {
                        'redirect_uri': CLIENT_LIST[client].get('REDIRECT_URL'),
                        'client_id': CLIENT_LIST[client].get('CLIENT_ID'),
                        'code': auth_code,
                        'code_verifier': code_verifier,
                        'grant_type': 'authorization_code'
                        }
                req = await self._session.post(
                    url=AUTH_REFRESH,
                    data = data,
                    headers=self._session_auth_headers.get(client),
                    allow_redirects=False
                    )
                token_data = await req.json()
            self._session_tokens[client] = {}
            for key in token_data:
                if '_token' in key:
                    self._session_tokens[client][key] = token_data[key]
            if 'error' in self._session_tokens[client]:
                errorTxt = self._session_tokens[client].get('error', '')
                if 'error_description' in self._session_tokens[client]:
                    error_description = self._session_tokens[client].get('error_description', '')
                    raise SeatException(f'{errorTxt} - {error_description}')
                else:
                    raise SeatException(errorTxt)
            if self._session_fulldebug:
                for key in self._session_tokens.get(client, {}):
                    if 'token' in key:
                        _LOGGER.debug(f'Got {key} for client {CLIENT_LIST[client].get("CLIENT_ID","")}, token: "{self._session_tokens.get(client, {}).get(key, None)}"')
            # Verify token, warn if problems are found
            verify = await self.verify_token(self._session_tokens[client].get('id_token', ''))
            if verify is False:
                _LOGGER.warning(f'Token for {client} is invalid!')
            elif verify is True:
                _LOGGER.debug(f'Token for {client} verified OK.')
            else:
                _LOGGER.warning(f'Token for {client} could not be verified, verification returned {verify}.')
            loop = asyncio.get_running_loop()
            rt = await loop.run_in_executor(None, self.writeTokenFile, client)
        except (SeatEULAException):
            _LOGGER.warning('Login failed, the terms and conditions might have been updated and need to be accepted. Login to  your local SEAT/Cupra site, e.g. "https://cupraofficial.se/" and accept the new terms before trying again')
            raise
        except (SeatAccountLockedException):
            _LOGGER.warning('Your account is locked, probably because of too many incorrect login attempts. Make sure that your account is not in use somewhere with incorrect password')
            raise
        except (SeatAuthenticationException):
            _LOGGER.warning('Invalid credentials or invalid configuration. Make sure you have entered the correct credentials')
            raise
        except (SeatException):
            _LOGGER.error('An API error was encountered during login, try again later')
            raise
        except (TypeError):
            _LOGGER.warning(self.anonymise(f'Login failed for {self._session_auth_username}. The server might be temporarily unavailable, try again later. If the problem persists, verify your account at your local SEAT/Cupra site, e.g. "https://cupraofficial.se/"'))
        except Exception as error:
            _LOGGER.error(self.anonymise(f'Login failed for {self._session_auth_username}, {error}'))
            return False
        return True

    async def _signin_service(self, html, authissuer, authorizationEndpoint, client=BRAND_CUPRA):
        """Method to signin to Connect portal."""
        # Extract login form and extract attributes
        try:
            response_data = await html.text()
            responseSoup = BeautifulSoup(response_data, 'html.parser')
            form_data = dict()
            if responseSoup is None:
                raise SeatLoginFailedException('Login failed, server did not return a login form')
            for t in responseSoup.find('form', id='emailPasswordForm').find_all('input', type='hidden'):
                if self._session_fulldebug:
                    _LOGGER.debug(f'Extracted form attribute: {t["name"], t["value"]}')
                form_data[t['name']] = t['value']
            form_data['email'] = self._session_auth_username
            pe_url = authissuer+responseSoup.find('form', id='emailPasswordForm').get('action')
        except Exception as e:
            _LOGGER.error('Failed to extract user login form.')
            raise

        # POST email
        self._session_auth_headers[client]['Referer'] = authorizationEndpoint
        self._session_auth_headers[client]['Origin'] = authissuer
        _LOGGER.debug(self.anonymise(f"Start authorization for user {self._session_auth_username}"))
        req = await self._session.post(
            url = pe_url,
            headers = self._session_auth_headers.get(client),
            data = form_data
        )
        if req.status != 200:
            raise SeatException('Authorization request failed')
        try:
            response_data = await req.text()
            responseSoup = BeautifulSoup(response_data, 'html.parser')
            pwform = {}
            credentials_form = responseSoup.find('form', id='credentialsForm')
            all_scripts = responseSoup.find_all('script', {'src': False})
            if credentials_form is not None:
                _LOGGER.debug('Found HTML credentials form, extracting attributes')
                for t in credentials_form.find_all('input', type='hidden'):
                    if self._session_fulldebug:
                        _LOGGER.debug(f'Extracted form attribute: {t["name"], t["value"]}')
                    pwform[t['name']] = t['value']
                    form_data = pwform
                    post_action = responseSoup.find('form', id='credentialsForm').get('action')
            elif all_scripts is not None:
                _LOGGER.debug('Found dynamic credentials form, extracting attributes')
                pattern = re.compile("templateModel: (.*?),\n")
                for sc in all_scripts:
                    if(pattern.search(sc.string)):
                        import json
                        data = pattern.search(sc.string)
                        jsondata = json.loads(data.groups()[0])
                        _LOGGER.debug(self.anonymise(f'JSON: {jsondata}'))
                        if not jsondata.get('hmac', False):
                            raise SeatLoginFailedException('Failed to extract login hmac attribute')
                        if not jsondata.get('postAction', False):
                            raise SeatLoginFailedException('Failed to extract login post action attribute')
                        if jsondata.get('error', None) is not None:
                            raise SeatLoginFailedException(f'Login failed with error: {jsondata.get("error", None)}')
                        form_data['hmac'] = jsondata.get('hmac', '')
                        post_action = jsondata.get('postAction')
            else:
                raise SeatLoginFailedException('Failed to extract login form data')
            form_data['password'] = self._session_auth_password
        except (SeatLoginFailedException) as e:
            raise
        except Exception as e:
            raise SeatAuthenticationException("Invalid username or service unavailable")

        # POST password
        self._session_auth_headers[client]['Referer'] = pe_url
        self._session_auth_headers[client]['Origin'] = authissuer
        _LOGGER.debug(f"Finalizing login")

        client_id = CLIENT_LIST[client].get('CLIENT_ID')
        pp_url = authissuer+'/'+post_action
        if not 'signin-service' in pp_url or not client_id in pp_url:
            pp_url = authissuer+'/signin-service/v1/'+client_id+"/"+post_action

        if self._session_fulldebug:
            _LOGGER.debug(f'Using login action url: "{pp_url}"')
        req = await self._session.post(
            url=pp_url,
            headers=self._session_auth_headers.get(client),
            data = form_data,
            allow_redirects=False
        )
        return req.headers.get('Location', None)

    async def terminate(self) -> None:
        """Log out from connect services"""
        for v in self.vehicles:
            _LOGGER.debug(self.anonymise(f'Calling stopFirebase() for vehicle {v.vin}'))
            newStatus = await v.stopFirebase()
            if newStatus != FIREBASE_STATUS_NOT_INITIALISED:
                _LOGGER.debug(self.anonymise(f'stopFirebase() not successful for vehicle {v.vin}'))
                # Although stopFirebase() was not successful, the firebase status is reset to FIREBASE_STATUS_NOT_INITIALISED to allow a new initialisation
                v.firebaseStatus = FIREBASE_STATUS_NOT_INITIALISED
        await self.logout()

    async def logout(self) -> None:
        """Logout, revoke tokens."""
        _LOGGER.info(f'Initiating logout.')
        self._session_headers.pop('Authorization', None)
        self._session_headers.pop('tokentype', None)
        self._session_headers['Content-Type'] = 'application/x-www-form-urlencoded'

        for client in self._session_tokens:
            # Ignore identity tokens
            for token_type in (
                token_type
                for token_type in self._session_tokens[client]
                if token_type in ['refresh_token', 'access_token']
            ):
                self._session_tokens[client][token_type] = None
        loop = asyncio.get_running_loop()
        await loop.run_in_executor(None, self.deleteTokenFile,)

  # HTTP methods to API
    async def get(self, url, vin=''):
        """Perform a HTTP GET."""
        try:
            response = await self._request(METH_GET, url)
            return response
        except aiohttp.client_exceptions.ClientResponseError as error:
            data = {
                'status_code': error.status,
                'error': error.code,
                'error_description': error.message,
                'response_headers': error.headers,
                'request_info': error.request_info
            }
            if error.status == 401:
                _LOGGER.warning('Received "Unauthorized" while fetching data. This can occur if tokens expired or refresh service is unavailable.')
                if self._error401 != True:
                    self._error401 = True
                    rc=await self.refresh_token(self._session_auth_brand)
                    if rc:
                        _LOGGER.info('Successfully refreshed tokens after error 401.')
                        self._error401 = False
                        #return True
                    else:
                        _LOGGER.info('Refresh of tokens after error 401 not successful.')
            elif error.status == 400:
                _LOGGER.error('Received "Bad Request" from server. The request might be malformed or not implemented correctly for this vehicle.')
            elif error.status == 412:
                _LOGGER.debug('Received "Pre-condition failed". Service might be temporarily unavailable.')
            elif error.status == 500:
                _LOGGER.info('Received "Internal server error". The service is temporarily unavailable.')
            elif error.status == 502:
                _LOGGER.info('Received "Bad gateway". Either the endpoint is temporarily unavailable or not supported for this vehicle.')
            elif 400 <= error.status <= 499:
                _LOGGER.error('Received unhandled error indicating client-side problem.\nRestart or try again later.')
            elif 500 <= error.status <= 599:
                _LOGGER.error('Received unhandled error indicating server-side problem.\nThe service might be temporarily unavailable.')
            else:
                _LOGGER.error('Received unhandled error while requesting API endpoint.')
            _LOGGER.debug(self.anonymise(f'HTTP request information: {data}'))
            return data
        except Exception as e:
            _LOGGER.debug(f'Got non HTTP related error: {e}')
            return {
                'error_description': 'Non HTTP related error'
            }

    async def post(self, url, **data):
        """Perform a HTTP POST."""
        if data:
            return await self._request(METH_POST, url, **data)
        else:
            return await self._request(METH_POST, url)

    async def _request(self, method, url, **kwargs):
        """Perform a HTTP query"""
        if self._session_fulldebug:
            argsString =''
            if len(kwargs)>0:
                argsString = 'with '
                for k, val in kwargs.items():
                    argsString = argsString + f"{k}=\'{val}\' " 
            _LOGGER.debug(self.anonymise(f'HTTP {method} "{url}" {argsString}'))
        try:
            if datetime.now(tz=None).date() != self._sessionRequestTimestamp.date():
                # A new day has begun. Store _sessionRequestCounter in history and reset timestamp and counter
                self._sessionRequestCounterHistory[self._sessionRequestTimestamp.strftime('%Y-%m-%d')]=self._sessionRequestCounter
                _LOGGER.info(f'History of the number of API calls:')
                for key, value in self._sessionRequestCounterHistory.items():
                    _LOGGER.info(f'   Date: {key}: {value} API calls')

                self._sessionRequestTimestamp= datetime.now(tz=None)
                self._sessionRequestCounter = 0
        except Exception as e:
            _LOGGER.error(f'Error while preparing output of API call history. Error: {e}')
        self._sessionRequestCounter = self._sessionRequestCounter + 1
        async with self._session.request(
            method,
            url,
            headers=self._session_headers if (PUBLIC_MODEL_IMAGES_SERVER not in url) else {}, # Set headers to {} when reading from PUBLIC_MODEL_IMAGES_SERVER
            timeout=ClientTimeout(total=TIMEOUT.seconds),
            cookies=self._session_cookies,
            raise_for_status=False,
            **kwargs
        ) as response:
            response.raise_for_status()

            # Update cookie jar
            if self._session_cookies != '':
                self._session_cookies.update(response.cookies)
            else:
                self._session_cookies = response.cookies

            try:
                if response.status == 204:
                    res = {'status_code': response.status}
                elif response.status == 202 and method==METH_PUT:
                    res = response
                elif response.status == 200 and method==METH_DELETE:
                    res = response
                elif response.status >= 200 or response.status <= 300:
                    # If this is a revoke token url, expect Content-Length 0 and return
                    if int(response.headers.get('Content-Length', 0)) == 0 and 'revoke' in url:
                        if response.status == 200:
                            return True
                        else:
                            return False
                    else:
                        if 'xml' in response.headers.get('Content-Type', ''):
                            res = xmltodict.parse(await response.text())
                        elif 'image/png'  in response.headers.get('Content-Type', ''):
                            res = await response.content.read()
                        else:
                            res = await response.json(loads=json_loads)
                else:
                    res = {}
                    _LOGGER.debug(self.anonymise(f'Not success status code [{response.status}] response: {response}'))
                if 'X-RateLimit-Remaining' in response.headers:
                    res['rate_limit_remaining'] = response.headers.get('X-RateLimit-Remaining', '')
            except Exception as e:
                res = {}
                _LOGGER.debug(self.anonymise(f'Something went wrong [{response.status}] response: {response}, error: {e}'))
                return res

            if self._session_fulldebug:
                if 'image/png'  in response.headers.get('Content-Type', ''):
                    _LOGGER.debug(self.anonymise(f'Request for "{url}" returned with status code [{response.status}]. Not showing response for Content-Type image/png.'))
                elif method==METH_PUT or method==METH_DELETE:
                    # deepcopy() of res can produce errors, if res is the API response on PUT or DELETE
                    _LOGGER.debug(f'Request for "{self.anonymise(url)}" returned with status code [{response.status}]. Not showing response for http {method}')
                else:
                    _LOGGER.debug(self.anonymise(f'Request for "{url}" returned with status code [{response.status}], response: {self.anonymise(deepcopy(res))}'))
            else:
                _LOGGER.debug(f'Request for "{url}" returned with status code [{response.status}]')
            return res

    async def _data_call(self, query, **data):
        """Function for POST actions with error handling."""
        try:
            response = await self.post(query, **data)
            _LOGGER.debug(self.anonymise(f'Data call returned: {response}'))
            return response
        except aiohttp.client_exceptions.ClientResponseError as error:
            _LOGGER.debug(self.anonymise(f'Request failed. Data: {data}, HTTP request headers: {self._session_headers}'))
            if error.status == 401:
                _LOGGER.error('Unauthorized')
            elif error.status == 400:
                _LOGGER.error(f'Bad request')
            elif error.status == 429:
                _LOGGER.warning('Too many requests. Further requests can only be made after the end of next trip in order to protect your vehicles battery.')
                return 429
            elif error.status == 500:
                _LOGGER.error('Internal server error, server might be temporarily unavailable')
            elif error.status == 502:
                _LOGGER.error('Bad gateway, this function may not be implemented for this vehicle')
            else:
                _LOGGER.error(f'Unhandled HTTP exception: {error}')
            #return False
        except Exception as error:
            _LOGGER.error(f'Failure to execute: {error}')
        return False

  # Class get data functions
    async def update_all(self) -> bool:
        """Update status."""
        try:
            # Get all Vehicle objects and update in parallell
            update_list = []
            for vehicle in self.vehicles:
                if vehicle.vin not in update_list:
                    _LOGGER.debug(self.anonymise(f'Adding {vehicle.vin} for data refresh'))
                    update_list.append(vehicle.update(updateType=1))
                else:
                    _LOGGER.debug(self.anonymise(f'VIN {vehicle.vin} is already queued for data refresh'))

            # Wait for all data updates to complete
            if len(update_list) == 0:
                _LOGGER.info('No vehicles in account to update')
            else:
                _LOGGER.debug('Calling update function for all vehicles')
                await asyncio.gather(*update_list)
            return True
        except (IOError, OSError, LookupError, Exception) as error:
            _LOGGER.warning(f'An error was encountered during interaction with the API: {error}')
        except:
            raise
        return False

    async def get_userData(self) -> dict:
        """Fetch user profile."""
        await self.set_token(self._session_auth_brand)
        userData={}
        #API_PERSONAL_DATA liefert fast das gleiche wie API_USER_INFO aber etwas weniger
        try:
            response = await self.get(eval(f"f'{API_PERSONAL_DATA}'"))
            if response.get('nickname'):
                userData= response
            else:
                _LOGGER.debug('Could not retrieve profile information')
        except:
            _LOGGER.debug('Could not fetch personal information.')

        try:
            response = await self.get(eval(f"f'{API_USER_INFO}'"))
            if response.get('name'):
                userData = response
            else:
                _LOGGER.debug('Could not retrieve profile information')
        except:
            _LOGGER.debug('Could not fetch personal information.')
        self._userData=userData
        return userData

    async def get_vehicles(self) -> list:
        """Fetch vehicle information from user profile."""
        api_vehicles = []
        # Check if user needs to update consent
        try:
            await self.set_token(self._session_auth_brand)
            #_LOGGER.debug('Achtung! getConsentInfo auskommentiert')
            response = await self.get(eval(f"f'{API_MBB_STATUSDATA}'"))
            if response.get('profileCompleted','incomplete'):
                if response.get('profileCompleted',False):
                    _LOGGER.debug('User consent is valid, no missing information for profile')
                else:
                    _LOGGER.debug('Profile incomplete. Please visit the web portal')
            else:
                _LOGGER.debug('Could not retrieve profile information')
            """consent = await self.getConsentInfo()
            if isinstance(consent, dict):
                _LOGGER.debug(f'Consent returned {consent}')
                if 'status' in consent.get('mandatoryConsentInfo', []):
                    if consent.get('mandatoryConsentInfo', [])['status'] != 'VALID':
                        _LOGGER.error(f'The user needs to update consent for {consent.get("mandatoryConsentInfo", [])["id"]}. If problems are encountered please visit the web portal first and accept terms and conditions.')
                elif len(consent.get('missingMandatoryFields', [])) > 0:
                    _LOGGER.error(f'Missing mandatory field for user: {consent.get("missingMandatoryFields", [])[0].get("name", "")}. If problems are encountered please visit the web portal first and accept terms and conditions.')
                else:
                    _LOGGER.debug('User consent is valid, no missing information for profile')
            else:
                _LOGGER.debug('Could not retrieve consent information')"""
        except:
            _LOGGER.debug('Could not fetch consent information. If problems are encountered please visit the web portal first and make sure that no new terms and conditions need to be accepted.')

        # Fetch vehicles
        try:
            legacy_vehicles = await self.get(eval(f"f'{API_VEHICLES}'"))
            if legacy_vehicles.get('vehicles', False):
                _LOGGER.debug('Found vehicle(s) associated with account.')
                for vehicle in legacy_vehicles.get('vehicles'):
                    vin = vehicle.get('vin', '')
                    self.addToAnonymisationDict(vin,'[VIN_ANONYMISED]')
                    response = await self.get(eval(f"f'{API_CAPABILITIES}'"))
                    #self._session_headers['Accept'] = 'application/json'
                    if response.get('capabilities', False):
                        vehicle["capabilities"]=response.get('capabilities')
                    else:
                        _LOGGER.warning(f"Failed to aquire capabilities information about vehicle with VIN {vehicle}.")
                        if vehicle.get('capabilities',None)!=None:
                            _LOGGER.warning(f"Keeping the old capability information.")
                        else:
                            _LOGGER.warning(f"Initialising vehicle without capabilities.")
                            vehicle["capabilities"]=[]
                    response = await self.get(eval(f"f'{API_CONNECTION}'"))
                    #self._session_headers['Accept'] = 'application/json'
                    if response.get('connection', False):
                        vehicle["connectivities"]=response.get('connection')
                    else:
                        _LOGGER.warning(f"Failed to aquire connection information about vehicle with VIN {vehicle}")
                    api_vehicles.append(vehicle)
        except:
            raise

        # If neither API returns any vehicles, raise an error
        if len(api_vehicles) == 0:
            raise SeatConfigException("No vehicles were found for given account!")
        # Get vehicle connectivity information
        else:
            try:
                for vehicle in api_vehicles:
                    _LOGGER.debug(self.anonymise(f'Checking vehicle {vehicle}'))
                    vin = vehicle.get('vin', '')
                    #for service in vehicle.get('connectivities', []):
                    #    if isinstance(service, str):
                    #        connectivity.append(service)
                    #    elif isinstance(service, dict):
                    #        connectivity.append(service.get('type', ''))

                    properties={}
                    for key in vehicle:
                        if not(key in {'capabilities', 'vin', 'specifications', 'connectivities'}):
                            properties[key]=vehicle.get(key)

                    newVehicle = {
                        'vin': vin,
                        'connectivities': vehicle.get('connectivities'),
                        'capabilities': vehicle.get('capabilities'),
                        'specification': vehicle.get('specifications'),
                        'properties': properties,
                    }
                    # Check if object already exist
                    _LOGGER.debug(f'Check if vehicle exists')
                    if self.vehicle(vin) is not None:
                        _LOGGER.debug(self.anonymise(f'Vehicle with VIN number {vin} already exist.'))
                        car = Vehicle(self, newVehicle)
                        if not car == self.vehicle(newVehicle):
                            _LOGGER.debug(self.anonymise(f'Updating {newVehicle} object'))
                            self._vehicles.pop(newVehicle)
                            self._vehicles.append(Vehicle(self, newVehicle))
                    else:
                        _LOGGER.debug(self.anonymise(f'Adding vehicle {vin}, with connectivities: {vehicle.get('connectivities')}'))
                        self._vehicles.append(Vehicle(self, newVehicle))
            except:
                raise SeatLoginFailedException("Unable to fetch associated vehicles for account")
        # Update data for all vehicles
        await self.update_all()

        return api_vehicles

 #### API get data functions ####
   # Profile related functions
    #async def getConsentInfo(self):
        """Get consent information for user."""
        """try:
            await self.set_token(self._session_auth_brand)
            atoken = self._session_tokens[self._session_auth_brand]['access_token']
            # Try old pyJWT syntax first
            try:
                subject = jwt.decode(atoken, verify=False).get('sub', None)
            except:
                subject = None
            # Try new pyJWT syntax if old fails
            if subject is None:
                try:
                    exp = jwt.decode(atoken, options={'verify_signature': False}).get('sub', None)
                    subject = exp
                except:
                    raise Exception("Could not extract sub attribute from token")

            data = {'scopeId': 'commonMandatoryFields'}
            response = await self.post(f'https://profileintegrityservice.apps.emea.vwapps.io/iaa/pic/v1/users/{subject}/check-profile', json=data)
            if response.get('mandatoryConsentInfo', False):
                data = {
                    'consentInfo': response
                }
                return data
            elif response.get('status_code', {}):
                _LOGGER.warning(f'Could not fetch realCarData, HTTP status code: {response.get("status_code")}')
            else:
                _LOGGER.info('Unhandled error while trying to fetch consent information')
        except Exception as error:
            _LOGGER.debug(f'Could not get consent information, error {error}')
        return False"""

    async def getBasicCarData(self, vin, baseurl) -> dict | bool:
        """Get car information from customer profile, VIN, nickname, etc."""
        await self.set_token(self._session_auth_brand)
        data={}
        try:
            response = await self.get(eval(f"f'{API_MYCAR}'"))
            if response.get('engines', {}):
                data['mycar']= response
            elif response.get('status_code', {}):
                _LOGGER.warning(f'Could not fetch vehicle mycar report, HTTP status code: {response.get("status_code")}')
            else:
                _LOGGER.info('Unhandled error while trying to fetch mycar data')
        except Exception as error:
            _LOGGER.warning(f'Could not fetch mycar report, error: {error}')
        if data=={}:
            return False
        return data

    async def getMileage(self, vin, baseurl) -> dict | bool:
        """Get car information from customer profile, VIN, nickname, etc."""
        await self.set_token(self._session_auth_brand)
        data={}
        try:
            response = await self.get(eval(f"f'{API_MILEAGE}'"))
            if response.get('mileageKm', {}):
                data['mileage'] = response
            elif response.get('status_code', {}):
                _LOGGER.warning(f'Could not fetch mileage information, HTTP status code: {response.get("status_code")}')
            else:
                _LOGGER.info('Unhandled error while trying to fetch mileage information')
        except Exception as error:
            _LOGGER.warning(f'Could not fetch mileage information, error: {error}')
        if data=={}:
            return False
        return data

    async def getVehicleHealthWarnings(self, vin, baseurl) -> dict | bool:
        """Get car information from customer profile, VIN, nickname, etc."""
        await self.set_token(self._session_auth_brand)
        data={}
        try:
            response = await self.get(eval(f"f'{API_WARNINGLIGHTS}'"))
            if 'statuses' in response:
                data['warninglights'] = response
            elif response.get('status_code', {}):
                _LOGGER.warning(f'Could not fetch warnlights, HTTP status code: {response.get("status_code")}')
            else:
                _LOGGER.info('Unhandled error while trying to fetch warnlights')
        except Exception as error:
            _LOGGER.warning(f'Could not fetch warnlights, error: {error}')
        if data=={}:
            return False
        return data

    #async def getOperationList(self, vin, baseurl):
        """Collect operationlist for VIN, supported/licensed functions."""
        """try:
            #await self.set_token('vwg')
            response = await self.get(f'{baseurl}/api/rolesrights/operationlist/v3/vehicles/{vin}')
            if response.get('operationList', False):
                data = response.get('operationList', {})
            elif response.get('status_code', {}):
                _LOGGER.warning(f'Could not fetch operation list, HTTP status code: {response.get("status_code")}')
                data = response
            else:
                _LOGGER.info(f'Could not fetch operation list: {response}')
                data = {'error': 'unknown'}
        except Exception as error:
            _LOGGER.warning(f'Could not fetch operation list, error: {error}')
            data = {'error': 'unknown'}
        return data"""

    async def getModelImageURL(self, vin, baseurl) -> dict | None:
        """Construct the URL for the model image."""
        await self.set_token(self._session_auth_brand)
        try:
            try:
                response = await self.get(
                    url=eval(f"f'{API_IMAGE}'"),
                )
                if response.get('front',False):
                    images: dict[str, str] ={}
                    for pos in {'front', 'side', 'top', 'rear'}:
                        if pos in response:
                            pic = await self._request(
                                METH_GET,
                                url=response.get(pos,False),
                            )
                            if len(pic)>0:
                                loop = asyncio.get_running_loop()
                                await loop.run_in_executor(None, self.writeImageFile, pos,pic, images)
                            if pos=='front':
                                # Crop the front image to a square format
                                try:
                                    im= Image.open(BytesIO(pic))
                                    width, height = im.size
                                    if height>width:
                                        width, height = height, width
                                    # Setting the points for cropped image
                                    left = (width-height)/2
                                    top = 0
                                    right = height+(width-height)/2
                                    bottom = height
                                    # Cropped image of above dimension
                                    im1 = im.crop((left, top, right, bottom))
                                    byteIO = BytesIO()
                                    im1.save(byteIO, format='PNG')
                                    await loop.run_in_executor(None, self.writeImageFile, pos+'_cropped',byteIO.getvalue(), images)
                                except:
                                    _LOGGER.warning('Cropping front image to square format failed.')
 
                    _LOGGER.debug('Read images from web site and wrote them to file.')
                    response['images']=images
                    return response
                else:
                    _LOGGER.debug(f'Could not fetch Model image URL, request returned with status code {response.status_code}')
            except:
                _LOGGER.debug('Could not fetch Model image URL')
        except:
            _LOGGER.debug('Could not fetch Model image URL, message signing failed.')
        return None

    async def getVehicleStatusReport(self, vin, baseurl) -> dict | bool:
        """Get stored vehicle status report (Connect services)."""
        data={}
        await self.set_token(self._session_auth_brand)
        try:
            response = await self.get(eval(f"f'{API_STATUS}'"))
            if response.get('doors', False):
                data['status']= response
            elif response.get('status_code', {}):
                _LOGGER.warning(f'Could not fetch vehicle status report, HTTP status code: {response.get("status_code")}')
            else:
                _LOGGER.info('Unhandled error while trying to fetch status data')
        except Exception as error:
            _LOGGER.warning(f'Could not fetch status report, error: {error}')
        if data=={}:
            return False
        return data

    async def getMaintenance(self, vin, baseurl) -> dict | bool:
        """Get stored vehicle status report (Connect services)."""
        data={}
        await self.set_token(self._session_auth_brand)
        try:
            response = await self.get(eval(f"f'{API_MAINTENANCE}'"))
            if response.get('inspectionDueDays', {}):
                data['maintenance'] = response
            elif response.get('status_code', {}):
                _LOGGER.warning(f'Could not fetch maintenance information, HTTP status code: {response.get("status_code")}')
            else:
                _LOGGER.info('Unhandled error while trying to fetch maintenance information')
        except Exception as error:
            _LOGGER.warning(f'Could not fetch maintenance information, error: {error}')
        if data=={}:
            return False
        return data

    async def getTripStatistics(self, vin, baseurl, supportsCyclicTrips) -> dict | bool:
        """Get short term and cyclic trip statistics."""
        await self.set_token(self._session_auth_brand)
        if self._session_tripStatisticsStartDate==None:
            # If connection was not initialised with parameter tripStatisticsStartDate, then 360 day is used for the CYCLIC trips and 90 days for the SHORT trips
            # (This keeps the statistics shorter in Home Assistant)
            startDate = (datetime.now() - timedelta(days= 360)).strftime('%Y-%m-%d')
        else:
            startDate = self._session_tripStatisticsStartDate
        try:
            data: dict[str, dict] ={'tripstatistics': {}} 
            if supportsCyclicTrips:
                dataType='CYCLIC'
                response = await self.get(eval(f"f'{API_TRIP}'"))
                if response.get('data', []):
                    data['tripstatistics']['cyclic']= response.get('data', [])
                elif response.get('status_code', {}):
                    _LOGGER.warning(f'Could not fetch trip statistics, HTTP status code: {response.get("status_code")}')
                else:
                    _LOGGER.info(f'Unhandled error while trying to fetch trip statistics')
            else:
                _LOGGER.info(f'Vehicle does not support cyclic trips.')
            dataType='SHORT'
            if self._session_tripStatisticsStartDate==None:
                # If connection was not initialised with parameter tripStatisticsStartDate, then 360 day is used for the CYCLIC trips and 90 days for the SHORT trips
                # (This keeps the statistics shorter in Home Assistant)
                startDate = (datetime.now() - timedelta(days= 90)).strftime('%Y-%m-%d')
            response = await self.get(eval(f"f'{API_TRIP}'"))
            if response.get('data', []):
                data['tripstatistics']['short']= response.get('data', [])
            elif response.get('status_code', {}):
                _LOGGER.warning(f'Could not fetch trip statistics, HTTP status code: {response.get("status_code")}')
            else:
                _LOGGER.info(f'Unhandled error while trying to fetch trip statistics')
            if data.get('tripstatistics',{}) != {}:
                return data
        except Exception as error:
            _LOGGER.warning(f'Could not fetch trip statistics, error: {error}')
        return False

    async def getPosition(self, vin, baseurl) -> dict | bool:
        """Get position data."""
        await self.set_token(self._session_auth_brand)
        try:
            response = await self.get(eval(f"f'{API_POSITION}'"))
            if response.get('lat', {}):
                data = {
                    'findCarResponse': response,
                    'isMoving': False
                }
                if hasattr(self, '_googleApiKey'):
                    apiKeyForGoogle= self._googleApiKey
                    lat= response.get('lat', 0)
                    lon= response.get('lon', 0)
                    response = await self.get(eval(f"f'{API_POS_TO_ADDRESS}'"))
                    if response.get('routes', []):
                        if response.get('routes', [])[0].get('legs', False):
                            data['findCarResponse']['position_to_address'] = response.get('routes', [])[0].get('legs',[])[0].get('start_address','')
                return data
            elif response.get('status_code', {}):
                if response.get('status_code', 0) == 204:
                    _LOGGER.debug(f'Seems car is moving, HTTP 204 received from position')
                    data = {
                        'isMoving': True,
                        'rate_limit_remaining': 15
                    }
                    return data
                else:
                    _LOGGER.warning(f'Could not fetch position, HTTP status code: {response.get("status_code")}')
            else:
                _LOGGER.info('Unhandled error while trying to fetch positional data')
        except Exception as error:
            _LOGGER.warning(f'Could not fetch position, error: {error}')
        return False

    async def getDeparturetimer(self, vin, baseurl) -> dict | bool:
        """Get departure timers."""
        await self.set_token(self._session_auth_brand)
        try:
            response = await self.get(eval(f"f'{API_DEPARTURE_TIMERS}'"))
            if response.get('timers', {}):
                data={}
                data['departureTimers'] = response
                return data
            elif response.get('status_code', {}):
                _LOGGER.warning(f'Could not fetch departure timers, HTTP status code: {response.get("status_code")}')
            else:
                _LOGGER.info('Unknown error while trying to fetch data for departure timers')
        except Exception as error:
            _LOGGER.warning(f'Could not fetch departure timers, error: {error}')
        return False

    async def getDepartureprofiles(self, vin, baseurl) -> dict | bool:
        """Get departure timers."""
        await self.set_token(self._session_auth_brand)
        try:
            response = await self.get(eval(f"f'{API_DEPARTURE_PROFILES}'"))
            if response.get('timers', {}):
                for e in range(len(response.get('timers', []))):
                    if response['timers'][e].get('singleTimer','')==None:
                        response['timers'][e].pop('singleTimer')
                    if response['timers'][e].get('recurringTimer','')==None:
                        response['timers'][e].pop('recurringTimer')
                data={}
                data['departureProfiles'] = response
                return data
            elif response.get('status_code', {}):
                _LOGGER.warning(f'Could not fetch departure profiles, HTTP status code: {response.get("status_code")}')
            else:
                _LOGGER.info('Unknown error while trying to fetch data for departure profiles')
        except Exception as error:
            _LOGGER.warning(f'Could not fetch departure profiles, error: {error}')
        return False

    async def getClimater(self, vin, baseurl, oldClimatingData) -> dict | bool:
        """Get climatisation data."""
        #data={}
        #data['climater']={}
        data = {'climater': oldClimatingData}
        await self.set_token(self._session_auth_brand)
        try:
            response = await self.get(eval(f"f'{API_CLIMATER_STATUS}'"))
            if response.get('climatisationStatus', {}):
                data['climater']['status']=response
            elif response.get('status_code', {}):
                _LOGGER.warning(f'Could not fetch climatisation status, HTTP status code: {response.get("status_code")}')
            else:
                _LOGGER.info('Unhandled error while trying to fetch climatisation status')
        except Exception as error:
            _LOGGER.warning(f'Could not fetch climatisation status, error: {error}')
        try:
            response = await self.get(eval(f"f'{API_CLIMATER}/settings'"))
            if response.get('targetTemperatureInCelsius', {}):
                data['climater']['settings']=response
            elif response.get('status_code', {}):
                _LOGGER.warning(f'Could not fetch climatisation settings, HTTP status code: {response.get("status_code")}')
            else:
                _LOGGER.info('Unhandled error while trying to fetch climatisation settings')
        except Exception as error:
            _LOGGER.warning(f'Could not fetch climatisation settings, error: {error}')
        if data['climater']=={}:
            return False
        return data

    async def getCharger(self, vin, baseurl, oldChargingData) -> dict | bool:
        """Get charger data."""
        await self.set_token(self._session_auth_brand)
        try:
            chargingStatus = {}
            chargingInfo = {}
            #chargingModes = {}
            #chargingProfiles = {}
            response = await self.get(eval(f"f'{API_CHARGING}/status'"))
            if response.get('battery', {}):
                chargingStatus = response
            elif response.get('status_code', {}):
                _LOGGER.warning(f'Could not fetch charging status, HTTP status code: {response.get("status_code")}')
            else:
                _LOGGER.info('Unhandled error while trying to fetch charging status')
            response = await self.get(eval(f"f'{API_CHARGING}/info'"))
            if response.get('settings', {}):
                chargingInfo = response
            elif response.get('status_code', {}):
                _LOGGER.warning(f'Could not fetch charging info, HTTP status code: {response.get("status_code")}')
            else:
                _LOGGER.info('Unhandled error while trying to fetch charging info')
            """response = await self.get(eval(f"f'{API_CHARGING}/modes'"))
            if response.get('battery', {}):
                chargingModes = response
            elif response.get('status_code', {}):
                _LOGGER.warning(f'Could not fetch charging modes, HTTP status code: {response.get("status_code")}')
            else:
                _LOGGER.info('Unhandled error while trying to fetch charging modes')"""
            """response = await self.get(eval(f"f'{API_CHARGING}/profiles'"))
            if response.get('profiles', {}):
                chargingProfiles = response
            elif response.get('status_code', {}):
                _LOGGER.warning(f'Could not fetch charging profiles, HTTP status code: {response.get("status_code")}')
            else:
                _LOGGER.info('Unhandled error while trying to fetch charging profiles')"""
            data = {'charging': oldChargingData}
            if chargingStatus != {}:
                data['charging']['status'] = chargingStatus
            else:
                _LOGGER.warning(f'getCharger() got no valid data for charging status')
            if chargingInfo != {}:
                data['charging']['info'] = chargingInfo
            else:
                _LOGGER.warning(f'getCharger() got no valid data for charging info')
            #if chargingModes != {}:
            #    data['charging']['modes'] = chargingModes
            #else:
            #    _LOGGER.warning(f'getCharger() got no valid data for charging modes')
            #if chargingProfiles != {}:
            #    data['charging']['profiles'] = chargingProfiles
            #else:
            #    _LOGGER.warning(f'getCharger() got no valid data for charging profiles')
            return data
        except Exception as error:
            _LOGGER.warning(f'Could not fetch charger, error: {error}')
        return False

    async def getPreHeater(self, vin, baseurl) -> dict | bool:
        """Get parking heater data."""
        await self.set_token(self._session_auth_brand)
        try:
            response = await self.get(eval(f"f'URL_not_yet_known'"))
            if response.get('statusResponse', {}):
                data = {'heating': response.get('statusResponse', {})}
                return data
            elif response.get('status_code', {}):
                _LOGGER.warning(f'Could not fetch pre-heating, HTTP status code: {response.get("status_code")}')
            else:
                _LOGGER.info('Unhandled error while trying to fetch pre-heating data')
        except Exception as error:
            _LOGGER.warning(f'Could not fetch pre-heating, error: {error}')
        return False

 #### API data set functions ####
    #async def get_request_status(self, vin, sectionId, requestId, baseurl):
        """Return status of a request ID for a given section ID."""
        """try:
            error_code = None
            # Requests for VW-Group API
            if sectionId == 'climatisation':
                capability='climatisation'
                url = eval(f"f'{API_REQUESTS}'")
            elif sectionId == 'batterycharge':
                capability='charging'
                url = eval(f"f'{API_REQUESTS}'")
            elif sectionId == 'departuretimer':
                capability='departure-timers'
                url = eval(f"f'{API_REQUESTS}'")
            elif sectionId == 'vsr':
                capability='status'
                url = eval(f"f'{API_REQUESTS}'")
            elif sectionId == 'rhf':
                capability='honkandflash'
                url = eval(f"f'{API_REQUESTS}/status'")
            else:
                capability='unknown'
                url = eval(f"f'{API_REQUESTS}'")

            response = await self.get(url)
            # Pre-heater on older cars
            if response.get('requestStatusResponse', {}).get('status', False):
                result = response.get('requestStatusResponse', {}).get('status', False)
            # Electric charging, climatisation and departure timers
            elif response.get('action', {}).get('actionState', False):
                result = response.get('action', {}).get('actionState', False)
                error_code = response.get('action', {}).get('errorCode', None)
            else:
                result = 'Unknown'
            # Translate status messages to meaningful info
            if result in ['request_in_progress', 'queued', 'fetched', 'InProgress', 'Waiting']:
                status = 'In progress'
            elif result in ['request_fail', 'failed']:
                status = 'Failed'
                if error_code is not None:
                    # Identified error code for charging, 11 = not connected
                    if sectionId == 'charging' and error_code == 11:
                        _LOGGER.info(f'Request failed, charger is not connected')
                    else:
                        _LOGGER.info(f'Request failed with error code: {error_code}')
            elif result in ['unfetched', 'delayed', 'PollingTimeout']:
                status = 'No response'
            elif result in [ "FailPlugDisconnected", "FailTimerChargingActive" ]:
                status = "Unavailable"
            elif result in ['request_successful', 'succeeded', "Successful"]:
                status = 'Success'
            else:
                status = result
            return status
        except Exception as error:
            _LOGGER.warning(f'Failure during get request status: {error}')
            raise SeatException(f'Failure during get request status: {error}')"""

    async def get_sec_token(self, spin, baseurl) -> str:
        """Get a security token, required for certain set functions."""
        data = {'spin': spin}
        url = eval(f"f'{API_SECTOKEN}'")
        response = await self.post(url, json=data, allow_redirects=True)
        if response.get('securityToken', False):
            return response['securityToken']
        else:
            raise SeatException('Did not receive a valid security token. Maybewrong SPIN?' )

    async def _setViaAPI(self, endpoint, **data) -> dict | bool:
        """Data call to API to set a value or to start an action."""
        await self.set_token(self._session_auth_brand)
        try:
            url = endpoint 
            response = await self._data_call(url, **data)
            if not response:
                raise SeatException(f'Invalid or no response for endpoint {endpoint}')
            elif response == 429:
                raise SeatThrottledException('Action rate limit reached. Start the car to reset the action limit')
            else:
                data = {'id': '', 'state' : ''}
                if 'requestId' in response:
                    data['state'] = 'Request accepted'
                for key in response:
                    if isinstance(response.get(key), dict):
                        for k in response.get(key):
                            if 'id' in k.lower():
                                data['id'] = str(response.get(key).get(k))
                            if 'state' in k.lower():
                                data['state'] = response.get(key).get(k)
                    else:
                        if 'Id' in key or 'id' in key:
                            data['id'] = str(response.get(key))
                        if 'State' in key:
                            data['state'] = response.get(key)
                if response.get('rate_limit_remaining', False):
                    data['rate_limit_remaining'] = response.get('rate_limit_remaining', None)
                return data
        except:
            raise
        return False

    async def _setViaPUTtoAPI(self, endpoint, **data) -> dict | bool:
        """PUT call to API to set a value or to start an action."""
        await self.set_token(self._session_auth_brand)
        try:
            url = endpoint 
            response = await self._request(METH_PUT,url, **data)
            if not response:
                raise SeatException(f'Invalid or no response for endpoint {endpoint}')
            elif response == 429:
                raise SeatThrottledException('Action rate limit reached. Start the car to reset the action limit')
            else:
                data = {'id': '', 'state' : ''}
                if 'requestId' in response:
                    data['state'] = 'Request accepted'
                for key in response:
                    if isinstance(response.get(key), dict):
                        for k in response.get(key):
                            if 'id' in k.lower():
                                data['id'] = str(response.get(key).get(k))
                            if 'state' in k.lower():
                                data['state'] = response.get(key).get(k)
                    else:
                        if 'Id' in key:
                            data['id'] = str(response.get(key))
                        if 'State' in key:
                            data['state'] = response.get(key)
                if response.get('rate_limit_remaining', False):
                    data['rate_limit_remaining'] = response.get('rate_limit_remaining', None)
                return data
        except:
            raise
        return False

    async def subscribe(self, vin, credentials) -> dict | bool:
        url = f'{APP_URI}/v2/subscriptions'
        deviceId = credentials.get('gcm',{}).get('app_id','')
        token = credentials.get('fcm',{}).get('registration',{}).get('token','')

        data = {
            "deviceId": deviceId, 
            "locale":"en_GB",
            "services":{"charging":True,"climatisation":True},
            "token": token, 
            "userId": self._user_id, 
            "vin":vin
            }
        return await self._setViaAPI(url, json=data)

    async def deleteSubscription(self, credentials):
        await self.set_token(self._session_auth_brand)
        try:
            id = credentials.get('subscription',{}).get('id','')
            url = f'{APP_URI}/v1/subscriptions/{id}'
            response = await self._request(METH_DELETE, url)
            if response.status==200: 
                _LOGGER.debug(f'Subscription {id} successfully deleted.')
                return response
            else:
                _LOGGER.debug(f'API did not successfully delete subscription.')
                raise SeatException(f'Invalid or no response for endpoint {url}')
                return response
        except aiohttp.client_exceptions.ClientResponseError as error:
            _LOGGER.debug(f'Request failed. Id: {id}, HTTP request headers: {self._session_headers}')
            if error.status == 401:
                _LOGGER.error('Unauthorized')
            elif error.status == 400:
                _LOGGER.error(f'Bad request')
            elif error.status == 429:
                _LOGGER.warning('Too many requests. Further requests can only be made after the end of next trip in order to protect your vehicles battery.')
                return 429
            elif error.status == 500:
                _LOGGER.error('Internal server error, server might be temporarily unavailable')
            elif error.status == 502:
                _LOGGER.error('Bad gateway, this function may not be implemented for this vehicle')
            else:
                _LOGGER.error(f'Unhandled HTTP exception: {error}')
            #return False
        except Exception as error:
            _LOGGER.error(f'Error: {error}')
            raise
        return False

    async def setCharger(self, vin, baseurl, mode, data) -> dict | bool:
        """Start/Stop charger."""
        if mode in {'start', 'stop'}:
            capability='charging'
            return await self._setViaAPI(eval(f"f'{API_REQUESTS}/{mode}'"))
        elif mode=='settings':
            return await self._setViaAPI(eval(f"f'{API_CHARGING}/{mode}'"), json=data)
        else:
            _LOGGER.error(f'Not yet implemented. Mode: {mode}. Command ignored')
            raise

    async def setClimater(self, vin, baseurl, mode, data, spin) -> dict | bool:
        """Execute climatisation actions."""
        try:
            # Only get security token if auxiliary heater is to be started
            if data.get('action', {}).get('settings', {}).get('heaterSource', None) == 'auxiliary':
                _LOGGER.error(f'This action is not yet implemented: {data.get('action', {}).get('settings', {}).get('heaterSource', None)}. Command ignored')
                #self._session_headers['X-securityToken'] = await self.get_sec_token(vin=vin, spin=spin, action='rclima', baseurl=baseurl)
                pass
            if mode == "stop": # Stop climatisation
                capability='climatisation'
                return await self._setViaAPI(eval(f"f'{API_REQUESTS}/stop'"))
            elif mode == "settings": # Set target temperature
                capability='climatisation'
                return await self._setViaAPI(eval(f"f'{API_CLIMATER}/settings'"), json=data)
            elif mode == "windowHeater stop": # Stop window heater
                capability='windowheating'
                return await self._setViaAPI(eval(f"f'{API_REQUESTS}/stop'"))
            elif mode == "windowHeater start": # Stop window heater
                capability='windowheating'
                return await self._setViaAPI(eval(f"f'{API_REQUESTS}/start'"))
            elif mode == "start": # Start climatisation
                return await self._setViaAPI(eval(f"f'{API_CLIMATER}/start'"), json = data)
            else: # Unknown modes
                _LOGGER.error(f'Unbekannter setClimater mode: {mode}. Command ignored')
                return False
        except:
            raise
        return False

    async def setDeparturetimer(self, vin, baseurl, data, spin) -> dict | bool:
        """Set departure timers."""
        try:
            url= eval(f"f'{API_DEPARTURE_TIMERS}'")
            if data:
                if data.get('minSocPercentage',False):
                    url=url+'/settings'
            return await self._setViaAPI(url, json = data)
        except:
            raise
        return False

    async def setDepartureprofile(self, vin, baseurl, data, spin) -> dict | bool:
        """Set departure profiles."""
        try:
            url= eval(f"f'{API_DEPARTURE_PROFILES}'")
            #if data:
                #if data.get('minSocPercentage',False):
                #    url=url+'/settings'
            return await self._setViaPUTtoAPI(url, json = data)
        except:
            raise
        return False

    async def sendDestination(self, vin, baseurl, data, spin):
        """Send destination to vehicle."""

        await self.set_token(self._session_auth_brand)
        try:
            url= eval(f"f'{API_DESTINATION}'")
            response = await self._request(METH_PUT, url, json=data)
            if response.status==202: #[202 Accepted]
                _LOGGER.debug(f'Destination {data[0]} successfully sent to API.')
                return response
            else:
                _LOGGER.debug(f'API did not successfully receive destination.')
                raise SeatException(f'Invalid or no response for endpoint {url}')
                return response
        except aiohttp.client_exceptions.ClientResponseError as error:
            _LOGGER.debug(f'Request failed. Data: {data}, HTTP request headers: {self._session_headers}')
            if error.status == 401:
                _LOGGER.error('Unauthorized')
            elif error.status == 400:
                _LOGGER.error(f'Bad request')
            elif error.status == 429:
                _LOGGER.warning('Too many requests. Further requests can only be made after the end of next trip in order to protect your vehicles battery.')
                return 429
            elif error.status == 500:
                _LOGGER.error('Internal server error, server might be temporarily unavailable')
            elif error.status == 502:
                _LOGGER.error('Bad gateway, this function may not be implemented for this vehicle')
            else:
                _LOGGER.error(f'Unhandled HTTP exception: {error}')
            #return False
        except Exception as error:
            _LOGGER.error(f'Error: {error}')
            raise
        return False

    async def setHonkAndFlash(self, vin, baseurl, data) -> dict | bool:
        """Execute honk and flash actions."""
        return await self._setViaAPI(eval(f"f'{API_HONK_AND_FLASH}'"), json = data)

    async def setLock(self, vin, baseurl, action, spin) -> dict | bool:
        """Remote lock and unlock actions."""
        try:
            # Fetch security token 
            self._session_headers['SecToken']= await self.get_sec_token(spin=spin, baseurl=baseurl)

            response = await self._setViaAPI(eval(f"f'{API_ACCESS}'"))

            # Clean up headers
            self._session_headers.pop('SecToken')

            return response

        except:
            self._session_headers.pop('SecToken')
            raise
        return False

    async def setPreHeater(self, vin, baseurl, data, spin) -> dict | bool:
        """Petrol/diesel parking heater actions."""
        try:
            # Fetch security token 
            self._session_headers['SecToken']= await self.get_sec_token(spin=spin, baseurl=baseurl)

            response = await self._setViaAPI(eval(f"f'url_not_yet_known'"), json = data)

            # Clean up headers
            self._session_headers.pop('SecToken')

            return response

        except:
            self._session_headers.pop('SecToken')
            raise
        return False

    async def setRefresh(self, vin, baseurl) -> dict | bool:
        """"Force vehicle data update."""
        return await self._setViaAPI(eval(f"f'{API_REFRESH}'"))

 #### Token handling ####
    async def validate_token(self, token) -> datetime:
        """Function to validate a single token."""
        try:
            now = datetime.now()
            # Try old pyJWT syntax first
            try:
                exp = jwt.decode(token, verify=False).get('exp', None)
            except:
                exp = None
            # Try new pyJWT syntax if old fails
            if exp is None:
                try:
                    exp = jwt.decode(token, options={'verify_signature': False}).get('exp', None)
                except:
                    raise Exception("Could not extract exp attribute")

            expires = datetime.fromtimestamp(int(exp))

            # Lazy check but it's very inprobable that the token expires the very second we want to use it
            if expires > now:
                return expires
            else:
                _LOGGER.debug(f'Token expired at {expires.strftime("%Y-%m-%d %H:%M:%S")}')
                return datetime.min # Return value datetime.min means that the token is not valid
        except Exception as e:
            _LOGGER.info(f'Token validation failed, {e}')
        return datetime.min # Return value datetime.min means that the token is not valid

    async def verify_token(self, token) -> bool:
        """Function to verify a single token."""
        try:
            req = None
            # Try old pyJWT syntax first
            try:
                aud = jwt.decode(token, verify=False).get('aud', None)
            except:
                aud = None
            # Try new pyJWT syntax if old fails
            if aud is None:
                try:
                    aud = jwt.decode(token, options={'verify_signature': False}).get('aud', None)
                except:
                    raise Exception("Could not extract exp attribute")

            if not isinstance(aud, str):
                aud = next(iter(aud))
            _LOGGER.debug(f"Verifying token for {aud}")
            # If audience indicates a client from https://identity.vwgroup.io
            for client in CLIENT_LIST:
                if self._session_fulldebug:
                    _LOGGER.debug(f"Matching {aud} against {CLIENT_LIST[client].get('CLIENT_ID', '')}")
                if aud == CLIENT_LIST[client].get('CLIENT_ID', ''):
                    req = await self._session.get(url = AUTH_TOKENKEYS)
                    break
            if req == None:
                return False
            
            # Fetch key list
            keys = await req.json()
            pubkeys = {}
            # Convert all RSA keys and store in list
            for jwk in keys['keys']:
                kid = jwk['kid']
                if jwk['kty'] == 'RSA':
                    pubkeys[kid] = jwt.algorithms.RSAAlgorithm.from_jwk(to_json(jwk))
            # Get key ID from token and get match from key list
            token_kid = jwt.get_unverified_header(token)['kid']
            if self._session_fulldebug:
                try:
                    _LOGGER.debug(f'Token Key ID is {token_kid}, match from public keys: {keys["keys"][token_kid]}')
                except:
                    pass
            pubkey = pubkeys[token_kid]

            # Verify token with public key
            if jwt.decode(token, key=pubkey, algorithms=['RS256'], audience=aud):
                return True
        except ExpiredSignatureError:
            return False
        except Exception as error:
            _LOGGER.debug(f'Failed to verify {aud} token, error: {error}')
        return False

    async def refresh_token(self, client) -> bool:
        """Function to refresh tokens for a client."""
        try:
            # Refresh API tokens
            _LOGGER.debug(f'Refreshing tokens for client "{client}"')
            if client != 'vwg':
                body = {
                    'grant_type': 'refresh_token',
                    'client_id': CLIENT_LIST[client].get('CLIENT_ID'),
                    'client_secret': CLIENT_LIST[client].get('CLIENT_SECRET'),
                    'refresh_token': self._session_tokens[client]['refresh_token']
                }
                url = AUTH_REFRESH
                self._session_token_headers[client]['User-ID']=self._user_id
                self._session_headers['User-ID']=self._user_id
            else:
                _LOGGER.error(f'refresh_token() does not support client \'{client}\' ')
                raise
            #    body = {
            #        'grant_type': 'refresh_token',
            #        'scope': 'sc2:fal',
            #        'token': self._session_tokens[client]['refresh_token']
            #    }
            #    url = 'https://mbboauth-1d.prd.ece.vwg-connect.com/mbbcoauth/mobile/oauth2/v1/token'

            try:
                response = await self._session.post(
                    url=url,
                    headers=self._session_token_headers.get(client),
                    data = body,
                )
            except:
                raise

            if response.status == 200:
                tokens = await response.json()
                # Verify access_token
                if 'access_token' in tokens:
                    if not await self.verify_token(tokens['access_token']):
                        _LOGGER.warning('Tokens could not be verified!')
                for token in tokens:
                    self._session_tokens[client][token] = tokens[token]
                return True
            elif response.status == 400:
                error = await response.json()
                if error.get('error', {}) == 'invalid_grant':
                    _LOGGER.debug(f'VW-Group API token refresh failed: {error.get("error_description", {})}')
                    #if client == 'vwg':
                    #    return await self._getAPITokens()
            else:
                resp = await response.json()
                _LOGGER.warning(f'Something went wrong when refreshing tokens for "{client}".')
                _LOGGER.debug(f'Headers: {TOKEN_HEADERS.get(client)}')
                _LOGGER.debug(f'Request Body: {body}')
                _LOGGER.warning(f'Something went wrong when refreshing VW-Group API tokens.')
        except Exception as error:
            _LOGGER.warning(f'Could not refresh tokens: {error}')
        return False

    async def set_token(self, client) -> bool:
        """Switch between tokens."""
        # Lock to prevent multiple instances updating tokens simultaneously
        async with self._lock:
            # If no tokens are available for client, try to authorize
            tokens = self._session_tokens.get(client, None)
            if tokens is None:
                _LOGGER.debug(f'Client "{client}" token is missing, call to authorize the client.')
                try:
                    # Try to authorize client and get tokens
                    if client != 'vwg':
                        result = await self._authorize(client)
                    else:
                        _LOGGER.error('getAPITokens() commented out.')
                        result = False
                        #result = await self._getAPITokens()

                    # If authorization wasn't successful
                    if result is not True:
                        raise SeatAuthenticationException(f'Failed to authorize client {client}')
                except:
                    raise
            try:
                # Validate access token for client, refresh if validation fails
                valid = await self.validate_token(self._session_tokens.get(client, {}).get('access_token', ''))
                if valid == datetime.min:
                    _LOGGER.debug(f'Tokens for "{client}" are invalid')
                    # Try to refresh tokens for client
                    if await self.refresh_token(client) is not True:
                        raise SeatTokenExpiredException(f'Tokens for client {client} are invalid')
                    else:
                        _LOGGER.debug(f'Tokens refreshed successfully for client "{client}"')
                        pass
                else:
                    try:
                        #dt = datetime.fromtimestamp(valid)
                        #_LOGGER.debug(f'Access token for "{client}" is valid until {dt.strftime("%Y-%m-%d %H:%M:%S")}')
                        _LOGGER.debug(f'Access token for "{client}" is valid until {valid.strftime("%Y-%m-%d %H:%M:%S")}')
                    except:
                        pass
                # Assign token to authorization header
                self._session_headers['Authorization'] = 'Bearer ' + self._session_tokens[client]['access_token']
            except:
                raise SeatException(f'Failed to set token for "{client}"')
            return True

 #### Class helpers ####
    @property
    def vehicles(self) -> list:
        """Return list of Vehicle objects."""
        return self._vehicles

    def vehicle(self, vin) -> Any:
        """Return vehicle object for given vin."""
        return next(
            (
                vehicle
                for vehicle in self.vehicles
                if vehicle.unique_id.lower() == vin.lower()
            ), None
        )

    def hash_spin(self, challenge, spin) -> str:
        """Convert SPIN and challenge to hash."""
        spinArray = bytearray.fromhex(spin);
        byteChallenge = bytearray.fromhex(challenge);
        spinArray.extend(byteChallenge)
        return hashlib.sha512(spinArray).hexdigest()

    def addToAnonymisationDict(self, keyword, replacement) -> None:
        self._anonymisationDict[keyword] = replacement

    def addToAnonymisationKeys(self, keyword) -> None:
        self._anonymisationKeys.add(keyword)

    def anonymise(self, inObj) -> Any:
        if self._session_anonymise:
            if isinstance(inObj, str):
                for key, value in self._anonymisationDict.items():
                    inObj = inObj.replace(key,value)
            elif isinstance(inObj, dict):
                for elem in inObj:
                    if elem in self._anonymisationKeys:
                        inObj[elem] = '[ANONYMISED]'
                    else:
                        inObj[elem]= self.anonymise(inObj[elem])
            elif isinstance(inObj, list):
                for i in range(len(inObj)):
                    inObj[i]= self.anonymise(inObj[i])
        return inObj

#async def main():
    """Main method."""
    """if '-v' in argv:
        logging.basicConfig(level=logging.INFO)
    elif '-vv' in argv:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.ERROR)

    async with ClientSession(headers={'Connection': 'keep-alive'}) as session:
        connection = Connection(session, **read_config())
        if await connection.doLogin():
            if await connection.get_vehicles():
                for vehicle in connection.vehicles:
                    print(f'Vehicle id: {vehicle}')
                    print('Supported sensors:')
                    for instrument in vehicle.dashboard().instruments:
                        print(f' - {instrument.name} (domain:{instrument.component}) - {instrument.str_state}')


if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())
"""